import copy
from typing import Union
import networkx as nx
from matplotlib import pyplot as plt
from .steps_base import *
from .preproc_steps import PreprocStepType, ImageToImageStep, TACsFromSegmentationStep, ResampleBloodTACStep
from .kinetic_modeling_steps import KMStepType, GraphicalAnalysisStep, TCMFittingAnalysisStep, ParametricGraphicalAnalysisStep

StepType = Union[FunctionBasedStep, ObjectBasedStep, PreprocStepType, KMStepType]

class StepsContainer:
    """
    A container for managing and executing a sequence of steps in a pipeline.

    This class allows for the addition, removal, and execution of steps, as well as printing their
    details and combining multiple step containers.

    Attributes:
        name (str): Name of the steps container.
        step_objs (list[StepType]): List of step objects in the container.
        step_names (list[str]): List of step names in the container.
    """
    def __init__(self, name: str, *steps: StepType):
        """
        Initializes the StepsContainer with a name and an optional sequence of steps.

        Args:
            name (str): Name of the steps container.
            *steps (StepType): Optional sequence of steps to add to the container.
        """
        self.name = name
        self.step_objs: list[StepType] = []
        self.step_names: list[str] = []
        for step in steps:
            self.add_step(step)
    
    def __repr__(self):
        """
        Provides an unambiguous string representation of the TACsFromSegmentationStep instance.

        Returns:
            str: A string representation showing how the instance can be recreated.
        """
        cls_name = type(self).__name__
        info_str = [f'{cls_name}(', f'{repr(self.name)},']
        
        for step_obj in self.step_objs:
            info_str.append(f'{repr(step_obj)},')
        
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    def add_step(self, step: StepType):
        """
        Adds a step to the container if it is not already present.

        Args:
            step (StepType): The step to add.

        Raises:
            TypeError: If the step is not of the correct type.
            KeyError: If a step with the same name already exists.
        """
        if not isinstance(step, StepType.__args__):
            raise TypeError("Step must be of type StepType")
        
        if step.name not in self.step_names:
            self.step_objs.append(copy.deepcopy(step))
            self.step_names.append(self.step_objs[-1].name)
        else:
            raise KeyError("A step with this name already exists.")
    
    def remove_step(self, step: Union[int, str]):
        """
        Removes a step from the container by index or name.

        Args:
            step (Union[int, str]): The index or name of the step to remove.

        Raises:
            IndexError: If the step index does not exist.
            KeyError: If the step name does not exist.
        """
        if isinstance(step, int):
            try:
                del self.step_objs[step]
                del self.step_names[step]
            except IndexError:
                raise IndexError(f"Step number {step} does not exist.")
        elif isinstance(step, str):
            if step not in self.step_names:
                raise KeyError(f"Step name {step} does not exist.")
            try:
                step_index = self.step_names.index(step)
                del self.step_objs[step_index]
                del self.step_names[step_index]
            except Exception:
                raise Exception
    
    def print_step_details(self):
        """
        Prints the details of all steps in the container.
        """
        print(f"({self.name} StepsContainer Info):")
        if not self.step_objs:
            print("No steps in container.")
        else:
            print("*" * 80)
            for step_id, a_step in enumerate(self.step_objs):
                print('-' * 70)
                print(f"Step Number {step_id + 1:>02}")
                print("  "+"    \n".join(str(a_step).split('\n')))
                print('-' * 70)
            print("*" * 80)
    
    def __str__(self) -> str:
        """
        Returns a detailed string representation of the StepsContainer instance.

        Returns:
            str: A string describing the steps-container, including its name, and the
            names of the sequence of steps.
        """
        info_str = [f"({self.name} steps-container info):"]
        if not self.step_objs:
            info_str.append('No steps in container.')
        else:
            info_str.append('-' * 70)
            for step_id, step_name in enumerate(self.step_names):
                info_str.append(f"  Step number {step_id + 1:>02}: {step_name}")
            info_str.append('-' * 70)
        return "\n".join(info_str)
    
    def __call__(self):
        """
        Executes all steps in the container in sequence.
        """
        for step_id, (step_name, a_step) in enumerate(zip(self.step_names, self.step_objs)):
            a_step()
    
    def __getitem__(self, step: Union[int, str]):
        """
        Gets a step from the container by index or name.

        Args:
            step (Union[int, str]): The index or name of the step to get.

        Returns:
            StepType: The requested step.

        Raises:
            IndexError: If the step index does not exist.
            KeyError: If the step name does not exist.
            TypeError: If the key is not an integer or string.
        """
        if isinstance(step, int):
            try:
                return self.step_objs[step]
            except IndexError:
                raise IndexError(f"Step number {step} does not exist.")
        elif isinstance(step, str):
            if step not in self.step_names:
                raise KeyError(f"Step name {step} does not exist.")
            try:
                step_index = self.step_names.index(step)
                return self.step_objs[step_index]
            except KeyError:
                raise KeyError(f"Step name {step} does not exist.")
        else:
            raise TypeError(f"Key must be an integer or a string. Got {type(step)}")
    
    def __add__(self, other: 'StepsContainer') -> 'StepsContainer':
        """
        Combines this StepsContainer with another one. The other container cannot have steps with the same name.

        Args:
            other (StepsContainer): The other steps container to combine with.

        Returns:
            StepsContainer: A new StepsContainer containing steps from both containers.

        Raises:
            TypeError: If the other object is not a StepsContainer.
            KeyError: If a step with the same name already exists.
        """
        if isinstance(other, StepsContainer):
            new_container_name = f"{self.name}-{other.name}"
            new_container = StepsContainer(new_container_name)
            for step in self.step_objs:
                new_container.add_step(step)
            for step in other.step_objs:
                new_container.add_step(step)
            return new_container
        else:
            raise TypeError("Can only add another StepsContainer instance")
    
    @classmethod
    def default_preprocess_steps(cls, name: str = 'preproc'):
        """
        Creates a default StepsContainer with common preprocessing steps.
        
        We have the following steps in sequence:
            - :meth:`Threshold Based Cropping<petpal.pipelines.preproc_steps.ImageToImageStep.default_threshold_cropping>`.
            - :meth:`Motion correct frames brighter than mean-image<petpal.pipelines.preproc_steps.ImageToImageStep.default_moco_frames_above_mean>`.
            - :meth:`Register PET to Anatomical<petpal.pipelines.preproc_steps.ImageToImageStep.default_register_pet_to_t1>`.
            - :meth:`Write ROI TACs from segmentation<petpal.pipelines.preproc_steps.TACsFromSegmentationStep.default_write_tacs_from_segmentation_rois>`.
            - :meth:`Resample blood on scanner frame times<petpal.pipelines.preproc_steps.ResampleBloodTACStep.default_resample_blood_tac_on_scanner_times>`.

        Args:
            name (str, optional): Name of the steps container. Defaults to 'preproc'.

        Returns:
            StepsContainer: A new StepsContainer with default preprocessing steps.
        """
        obj = cls(name=name)
        obj.add_step(ImageToImageStep.default_threshold_cropping())
        obj.add_step(ImageToImageStep.default_windowed_moco())
        obj.add_step(ImageToImageStep.default_register_pet_to_t1())
        obj.add_step(TACsFromSegmentationStep.default_write_tacs_from_segmentation_rois())
        obj.add_step(ResampleBloodTACStep.default_resample_blood_tac_on_scanner_times())
        return obj
    
    @classmethod
    def default_graphical_analysis_steps(cls, name: str = 'km_graphical_analysis'):
        """
        Creates a default StepsContainer with common graphical analysis steps.
        
        We have the following steps in sequence:
            - :meth:`Patlak<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_patlak>`.
            - :meth:`Logan<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_logan>`.
            - :meth:`Alt-Logan<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_alt_logan>`.

        Args:
            name (str, optional): Name of the steps container. Defaults to 'km_graphical_analysis'.

        Returns:
            StepsContainer: A new StepsContainer with default graphical analysis steps.
            
        Notes:
            The steps do not technically depend on each other and can be run out of sequence.
        """
        obj = cls(name=name)
        obj.add_step(GraphicalAnalysisStep.default_patlak())
        obj.add_step(GraphicalAnalysisStep.default_logan())
        obj.add_step(GraphicalAnalysisStep.default_alt_logan())
        return obj
    
    @classmethod
    def default_parametric_graphical_analysis_steps(cls, name: str = 'km_parametric_graphical_analysis'):
        """
        Creates a default StepsContainer with common parametric graphical analysis steps.
        
        We have the following steps in sequence:
            - :meth:`Patlak<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_patlak>`.
            - :meth:`Logan<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_logan>`.
            - :meth:`Alt-Logan<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_alt_logan>`.

        Args:
            name (str, optional): Name of the steps container. Defaults to 'km_parametric_graphical_analysis'.

        Returns:
            StepsContainer: A new StepsContainer with default parametric graphical analysis steps.
            
        Notes:
            The steps do not technically depend on each other and can be run out of sequence.
        """
        obj = cls(name=name)
        obj.add_step(ParametricGraphicalAnalysisStep.default_patlak())
        obj.add_step(ParametricGraphicalAnalysisStep.default_logan())
        obj.add_step(ParametricGraphicalAnalysisStep.default_alt_logan())
        return obj
    
    @classmethod
    def default_tcm_analysis_steps(cls, name: str = 'km_tcm_analysis'):
        """
        Creates a default StepsContainer with common TCM analysis steps.
        
        We have the following steps in sequence:
            - :meth:`1TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_1tcm>`.
            - :meth:`Serial 2TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_serial2tcm>`.
            - :meth:`Irreversible 2TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_irreversible_2tcm>`.

        Args:
            name (str, optional): Name of the steps container. Defaults to 'km_tcm_analysis'.

        Returns:
            StepsContainer: A new StepsContainer with default TCM analysis steps.
        """
        obj = cls(name=name)
        obj.add_step(TCMFittingAnalysisStep.default_1tcm())
        obj.add_step(TCMFittingAnalysisStep.default_serial2tcm())
        obj.add_step(TCMFittingAnalysisStep.default_irreversible_2tcm())
        return obj
    
    @classmethod
    def default_kinetic_analysis_steps(cls, name: str = 'km'):
        """
        Creates a default StepsContainer with common kinetic analysis steps.
        
        We have the following steps in sequence:
            - :meth:`ROI TACs: Patlak<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_patlak>`.
            - :meth:`ROI TACs: Logan<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_logan>`.
            - :meth:`ROI TACs: Alt-Logan<petpal.pipelines.kinetic_modeling_steps.GraphicalAnalysisStep.default_alt_logan>`.
            - :meth:`Parametric: Patlak<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_patlak>`.
            - :meth:`Parametric: Logan<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_logan>`.
            - :meth:`Parametric: Alt-Logan<petpal.pipelines.kinetic_modeling_steps.ParametricGraphicalAnalysisStep.default_alt_logan>`.
            - :meth:`ROI TACs: 1TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_1tcm>`.
            - :meth:`ROI TACs: Serial 2TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_serial2tcm>`.
            - :meth:`ROI TACs: Irreversible 2TCM<petpal.pipelines.kinetic_modeling_steps.TCMFittingAnalysisStep.default_irreversible_2tcm>`.

        Args:
            name (str, optional): Name of the steps container. Defaults to 'km'.

        Returns:
            StepsContainer: A new StepsContainer with default kinetic analysis steps.
            
        Notes:
            The steps do not technically depend on each other and can be run out of sequence.
        """
        
        parametric_graphical_analysis_steps = cls.default_parametric_graphical_analysis_steps()
        graphical_analysis_steps = cls.default_graphical_analysis_steps()
        tcm_analysis_steps = cls.default_tcm_analysis_steps()
        
        obj = parametric_graphical_analysis_steps + graphical_analysis_steps + tcm_analysis_steps
        obj.name = name
        return obj


class StepsPipeline:
    """
    A pipeline for managing and executing a sequence of steps organized in containers, with support for dependencies.

    This class allows for the addition, removal, and execution of step containers,
    as well as managing dependencies between the steps.

    Attributes:
        name (str): Name of the steps pipeline.
        step_containers (dict[str, StepsContainer]): Dictionary of step containers in the pipeline.
        dependency_graph (nx.DiGraph): Directed acyclic graph representing dependencies between steps.
        
    See Also:
        :class:`StepsContainer`
    """
    def __init__(self, name: str, step_containers: list[StepsContainer]):
        """
        Initializes the StepsPipeline with a name and an optional list of step containers.

        Args:
            name (str): Name of the steps pipeline.
            step_containers (list[StepsContainer]): Optional list of step containers to add to the pipeline.
        """
        self.name: str = name
        self.step_containers: dict[str, StepsContainer] = {}
        self.dependency_graph = nx.DiGraph()
        
        for container in step_containers:
            self.add_container(container)
    
    def __repr__(self):
        """
        Provides an unambiguous string representation of the TACsFromSegmentationStep instance.

        .. important::
            This ``repr`` does not include the dependency graph.

        Returns:
            str: A string representation showing how the instance can be recreated.
        """
        cls_name = type(self).__name__
        info_str = [f'{cls_name}(', f'name={repr(self.name)},' 'step_containers=[']
        
        for _, container_obj in self.step_containers.items():
            info_str.append(f'{repr(container_obj)},')
        info_str.append(']')
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    def __str__(self) -> str:
        """
        Returns a detailed string representation of the StepsPipeline instance.

        Returns:
            str: A string describing the steps-pipeline, including its name, the steps in
            each of the steps-containers, and the output dependencies of each step.
        """
        sep_line_len = 50
        info_str = [f'({self.name} pipeline info):', "*" * sep_line_len]
        for con_name, con_obj in self.step_containers.items():
            info_str.append(f"  ({con_name} steps-container info):")
            for step_num, step_name in enumerate(con_obj.step_names):
                info_str.append(f"    Step {step_num:>02}: {repr(step_name)}")
                step_deps = list(self.dependency_graph[step_name].keys())
                if step_deps:
                    info_str.append(fr"      \> sends output to:")
                    for dep_name in step_deps[:-1]:
                        info_str.append(fr"            ├── {repr(dep_name)}")
                    for dep_name in step_deps[-1:]:
                        info_str.append(fr"            └── {repr(dep_name)}")
            info_str.append("-" * sep_line_len)
        info_str.append("*" * sep_line_len)
        return "\n".join(info_str)
    
    def add_container(self, step_container: StepsContainer):
        """
        Adds a step container to the pipeline.

        Args:
            step_container (StepsContainer): The step container to add.

        Raises:
            TypeError: If the step container is not an instance of StepsContainer.
            KeyError: If a container with the same name already exists.
        """
        if not isinstance(step_container, StepsContainer):
            raise TypeError("`step_container` must be an instance of StepsContainer")
        
        container_name = step_container.name
        if container_name in self.step_containers.keys():
            raise KeyError(f"Container name {container_name} already exists.")
        
        self.step_containers[container_name] = copy.deepcopy(step_container)
        
        step_objs = self.step_containers[container_name].step_objs
        for step in step_objs:
            self.dependency_graph.add_node(f"{step.name}", grp=container_name)
    
    def __call__(self):
        """
        Executes all steps in the pipeline in topologically sorted order.
        """
        for step_name in nx.topological_sort(self.dependency_graph):
            step = self.get_step_from_node_label(node_label=step_name)
            step()
    
    def add_step(self, container_name: str, step: StepType):
        """
        Adds a step to a specified container in the pipeline.

        Args:
            container_name (str): The name of the container to add the step to.
            step (StepType): The step to add.

        Raises:
            KeyError: If the container or step name does not exist or if the step name already exists.
        """
        if container_name not in self.step_containers.keys():
            raise KeyError(f"Container name {container_name} does not exist.")
        if step.name in self.dependency_graph:
            raise KeyError(f"Step name {step.name} already exists. Steps must have unique names.")
        
        self.step_containers[container_name].add_step(step)
        self.dependency_graph.add_node(f"{step.name}", grp=container_name)
    
    def remove_step(self, step: str):
        """
        Removes a step from the pipeline based on its name.

        Args:
            step (str): The name of the step to remove.

        Raises:
            KeyError: If the step name does not exist.
        """
        node_names = list(self.dependency_graph.nodes)
        if step not in node_names:
            raise KeyError(f"Step name {step} does not exist.")
        graph_nodes = self.dependency_graph.nodes(data=True)
        container_name = graph_nodes[step]['grp']
        self.dependency_graph.remove_node(step)
        self.step_containers[container_name].remove_step(step)
    
    def print_steps_names(self, container_name: Union[str, None] = None):
        """
        Prints the names of all steps in the pipeline or a specific container.

        Args:
            container_name (Union[str, None], optional): The name of the container to print step names from.
                If None, prints step names from all containers.

        Raises:
            KeyError: If the container name does not exist.
        """
        if container_name is None:
            for name, container in self.step_containers.items():
                print(str(container))
        elif container_name in self.step_containers.keys():
            print(str(self.step_containers[container_name]))
        else:
            raise KeyError(f"Container name {container_name} does not exist. ")
    
    def print_steps_details(self, container_name: Union[str, None] = None):
        """
        Prints the details of all steps in the pipeline or a specific container.

        Args:
            container_name (Union[str, None], optional): The name of the container to print step details from.
                If None, prints step details from all containers.

        Raises:
            KeyError: If the container name does not exist.
        """
        if container_name is None:
            for name, container in self.step_containers.items():
                container.print_step_details()
        elif container_name in self.step_containers.keys():
            self.step_containers[container_name].print_step_details()
        else:
            raise KeyError(f"Container name {container_name} does not exist. ")
    
    def add_dependency(self, sending: str, receiving: str):
        """
        Adds a dependency between two steps in the pipeline.

        Args:
            sending (str): The name of the sending step.
            receiving (str): The name of the receiving step.

        Raises:
            KeyError: If either step name does not exist.
            RuntimeError: If adding the dependency creates a cycle in the dependency graph.
        """
        node_names = list(self.dependency_graph.nodes)
        if sending not in node_names:
            raise KeyError(f"Step name {sending} does not exist.")
        if receiving not in node_names:
            raise KeyError(f"Step name {receiving} does not exist.")
        
        self.dependency_graph.add_edge(sending, receiving)
        
        if not nx.is_directed_acyclic_graph(self.dependency_graph):
            raise RuntimeError(f"Adding dependency {sending} -> {receiving} creates a cycle!")
    
    def get_step_from_node_label(self, node_label: str) -> StepType:
        """
        Retrieves a step object based on its node label in the dependency graph.

        Args:
            node_label (str): The label of the node representing the step.

        Returns:
            StepType: The step object.

        Raises:
            KeyError: If the step name does not exist in the graph.
        """
        if node_label not in self.dependency_graph.nodes:
            raise KeyError(f"Step name {node_label} does not exist.")
        graph_nodes = self.dependency_graph.nodes(data=True)
        container_name = graph_nodes[node_label]['grp']
        return self.step_containers[container_name][node_label]
    
    def update_dependencies_for(self, step_name: str, verbose=False):
        """
        Updates the dependencies for a specified step, setting inputs from the outputs of its dependencies.
        Gets the list of steps that are sending outputs to this step. The order of dependency may matter
        depending on the implementation of `set_input_as_output_from` for the step.

        Args:
            step_name (str): The name of the step to update dependencies for.
            verbose (bool, optional): If True, prints detailed information during the update process. Defaults to False.
        """
        input_edges = self.dependency_graph.in_edges(step_name)
        input_step_names = [edge[0] for edge in input_edges]
        if input_step_names:
            input_steps = (self.get_step_from_node_label(step_name) for step_name in input_step_names)
            receiving_step = self.get_step_from_node_label(step_name)
            receiving_step.set_input_as_output_from(*input_steps)
            if verbose:
                print(f"Step {step_name} received inputs from:\n{input_step_names}")



    def update_dependencies(self, verbose=False):
        """
        Updates the dependencies for all steps in the pipeline.

        Args:
            verbose (bool, optional): If True, prints detailed information during the update process. Defaults to False.
        """
        for step_name in nx.topological_sort(self.dependency_graph):
            self.update_dependencies_for(step_name=step_name, verbose=verbose)
    
    def get_steps_potential_run_state(self) -> dict:
        """
        Gets a dictionary representing the potential run state of all steps in topologically sorted order.

        Returns:
            dict: Dictionary mapping step names to their potential run state (True if the step can potentially run, False otherwise).
        """
        step_maybe_runnable = {}
        for node_name in nx.topological_sort(self.dependency_graph):
            step = self.get_step_from_node_label(node_name)
            step_maybe_runnable[node_name] = step.can_potentially_run()
        return step_maybe_runnable
    
    def can_steps_potentially_run(self) -> bool:
        """
        Checks if all steps in the pipeline can potentially run.

        Returns:
            bool: True if all steps can potentially run, False otherwise.
        """
        steps_run_state = self.get_steps_potential_run_state()
        for step_name, run_state in steps_run_state.items():
            if not run_state:
                return False
        return True
    
    def plot_dependency_graph(self,
                              figsize: tuple = (12, 12),
                              node_options: Union[dict, None] = None,
                              label_options: Union[dict, None] = None,
                              draw_options: Union[dict, None] = None) -> None:
        """
        Plots the dependency graph of the steps pipeline.

        Args:
            figsize (tuple, optional): Figure size for the plot. Defaults to (12, 12).
            node_options (dict, optional): Additional options for drawing nodes. Defaults to None.
            label_options (dict, optional): Additional options for node labels. Defaults to None.
            draw_options (dict, optional): Additional options for the draw_networkx function. Defaults to None.
        
        Notes:
            We have the following node options.
            
            
            .. code-block:: python
            
                default_node_options = {'node_size': 2000, 'node_color': 'w', 'node_shape': 'o'}
        
                default_label_options = {'verticalalignment': 'center',
                                         'clip_on': False,
                                         'font_size': 15,
                                         'bbox' : dict(facecolor='azure', edgecolor='blue',
                                                       boxstyle='round', alpha=0.25, linewidth=2)
                                        }
                
                default_draw_options = {'with_labels': True, 'arrows': True, 'arrowsize': 30, 'width': 2}
            
        """
        dep_graph = self.dependency_graph
        
        for layer, nodes in enumerate(nx.topological_generations(dep_graph)):
            for node in nodes:
                dep_graph.nodes[node]['layer'] = layer
        
        dep_pos = nx.multipartite_layout(dep_graph, subset_key='layer', align='horizontal')
        dep_labels = {a_lab: '\n'.join(a_lab.split("_")) for a_lab in dep_graph.nodes}
        
        default_node_options = {'node_size': 2000, 'node_color': 'w', 'node_shape': 'o'}
        
        default_label_options = {'verticalalignment': 'center',
                                 'clip_on': False,
                                 'font_size': 15,
                                 'bbox' : dict(facecolor='azure', edgecolor='blue',
                                               boxstyle='round', alpha=0.25, linewidth=2)
                                }
        
        default_draw_options = {'with_labels': True, 'arrows': True, 'arrowsize': 30, 'width': 2}
        
        if node_options:
            default_node_options = {**default_node_options, **node_options}
        if label_options:
            default_label_options = {**default_label_options, **label_options}
        if draw_options:
            default_draw_options = {**default_draw_options, **draw_options}
        
        myFig, myAx = plt.subplots(figsize=figsize, constrained_layout=True)
        nx.draw_networkx(dep_graph, ax=myAx, pos=dep_pos, labels=dep_labels,
                         **default_node_options,
                         **default_label_options,
                         **default_draw_options)
        
        plt.show()
    
    def print_dependency_graph(self):
        """
        Prints a textual representation of the dependency graph for the steps pipeline.

        This method outputs the steps in the pipeline in topologically sorted order,
        showing their dependencies. For each step, it lists the subsequent steps
        that depend on it, formatted in a tree-like structure.
        """
        sep_line_len = 50
        info_str = [f'({self.name} pipeline dependency info):', "*" * sep_line_len]
        for step_name in nx.topological_sort(self.dependency_graph):
            info_str.append(f'{repr(step_name)}')
            step_deps = list(self.dependency_graph[step_name].keys())
            if step_deps:
                info_str[-1] = f"{info_str[-1]} sends output to"
                for dep_name in step_deps[:-1]:
                    info_str.append(f"  ├── {repr(dep_name)}")
                for dep_name in step_deps[-1:]:
                    info_str.append(f"  └── {repr(dep_name)}")
            else:
                info_str[-1] = f"{info_str[-1]} has no output dependencies"
        
        info_str.append('*' * sep_line_len)
        print('\n'.join(info_str))
        
    
    @classmethod
    def default_steps_pipeline(cls, name='PET-MR_analysis'):
        """
        Creates a default steps pipeline with predefined containers and dependencies.
        
        Contains the following step-containers:
            - :meth:`Pre-processing<StepsContainer.default_preprocess_steps>`.
            - :meth:`Kinetic modeling<StepsContainer.default_kinetic_analysis_steps>`.
            
        The pipeline has the following steps dependency:
        
        .. plot::
            :include-source:
            :caption: The dependency graph of the steps pipeline. From raw PET `.nii` images we crop the images
                in each dimension based on a threshold, perform motion correction for each frame where the per-frame
                mean-intensity is larger than the overall time-averaged intensity of the PET series, register the
                PET series to anatomical space, extract all the TACs corresponding the ROIs defined via segmentation,
                and resample the blood TAC onto the scanner frame times as part of pre-processing. Then, we perform
                ROI-based graphical and full tissue compartment model fits, and parametric graphical analysis fits.
                
            from petpal.pipelines.steps_containers import StepsPipeline
            
            # Instantiate the default PET/MR pipeline
            my_pipe = StepsPipeline.default_steps_pipeline()
            
            # Plot the dependency graph of the steps pipeline
            my_pipe.plot_dependency_graph()
        

        Args:
            name (str, optional): Name of the steps pipeline. Defaults to 'PET-MR_analysis'.

        Returns:
            StepsPipeline: A new StepsPipeline instance with default setup.
        """
        obj = cls(name=name, step_containers=[StepsContainer.default_preprocess_steps(name='preproc'),
                                              StepsContainer.default_kinetic_analysis_steps(name='km')])
        
        obj.add_dependency(sending='thresh_crop', receiving='moco_frames_above_mean')
        obj.add_dependency(sending='moco_frames_above_mean', receiving='register_pet_to_t1')
        obj.add_dependency(sending='register_pet_to_t1', receiving='write_roi_tacs')
        obj.add_dependency(sending='register_pet_to_t1', receiving='resample_PTAC_on_scanner')
        
        for method in ['patlak', 'logan', 'alt_logan']:
            obj.add_dependency(sending='register_pet_to_t1', receiving=f'parametric_{method}_fit')
            obj.add_dependency(sending='resample_PTAC_on_scanner', receiving=f'parametric_{method}_fit')
        
        for fit_model in ['1tcm', '2tcm-k4zero', 'serial-2tcm', 'patlak', 'logan', 'alt_logan']:
            obj.add_dependency(sending='write_roi_tacs', receiving=f"roi_{fit_model}_fit")
            obj.add_dependency(sending='resample_PTAC_on_scanner', receiving=f"roi_{fit_model}_fit")
        
        return obj


"""
Command-line interface (CLI) for conducting graphical analysis of PET Time-Activity Curves (TACs).

This module provides a CLI to interact with the graphical_analysis module. It leverages argparse to handle command-line
arguments.

The user must provide:
    * Input TAC file path
    * Region of Interest (ROI) TAC file path
    * Threshold in minutes (below which data points will not be considered for fitting)
    * The method name for conducting the analysis. Supported methods are 'patlak', 'logan', or 'alt-logan'.
    * Output directory where the analysis results will be saved

An optional filename prefix for the output files can also be supplied.

This script utilizes the :class:`petpal.kinetic_modeling.graphical_analysis.GraphicalAnalysis` class to perform the graphical analysis
and save the results accordingly.

Example:
    .. code-block:: bash
    
        petpal-graph-analysis --input-tac-path /path/to/input.tac --roi-tac-path /path/to/roi.tac --threshold-in-mins 30.0 --method-name patlak --output-directory ./analysis --output-filename-prefix analysis

See Also:
    :mod:`petpal.kinetic_modeling.graphical_analysis` - module responsible for conducting and saving graphical analysis of PET TACs.

"""

import argparse
from ..kinetic_modeling import graphical_analysis as pet_ga


def _add_common_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("-i", "--input-tac-path", required=True, help="Path to the input TAC file.")
    parser.add_argument("-o", "--output-directory", required=True, help="Path to the output directory.")
    parser.add_argument("-p", "--output-filename-prefix", required=True, help="Prefix for the output filenames.")
    parser.add_argument("-t", "--threshold-in-mins", required=True, type=float,
                           help="Threshold in minutes for the analysis.")
    parser.add_argument("-m", "--method-name", required=True, choices=['patlak', 'logan', 'alt-logan', 'logan-ref'],
                           help="Analysis method to be used.")
    parser.add_argument("-k","--k2-prime",required=False,help="k2-prime value used for logan-ref only",type=float)
    parser.add_argument("--print", action="store_true", help="Whether to print the analysis results.",default=False)


def main():
    parser = argparse.ArgumentParser(prog="Graphical Analysis", description="Perform graphical analysis on TAC data.",
                                     epilog="Example: petpal-graph-analysis "
                                            "graphical-analysis "
                                            "--input-tac-path /path/to/input.tac "
                                            "--input-image-path /path/to/pet4D.img "
                                            "--output-directory /path/to/output "
                                            "--output-filename-prefix graph_ana "
                                            "--method-name patlak --threshold-in-mins 30.0 ")
    subparsers = parser.add_subparsers(dest="command", help="Sub-command help.")
    parser_single_roi = subparsers.add_parser('graphical-analysis')
    _add_common_args(parser_single_roi)
    parser_single_roi.add_argument("-r", "--roi-tac-path", required=True, help="Path to the ROI TAC file.")

    parser_multitac = subparsers.add_parser('graphical-analysis-multitac')
    _add_common_args(parser_multitac)
    parser_multitac.add_argument("-r", "--roi-tacs-dir", required=True, help="Path to directory containing ROI TTACs")

    args = parser.parse_args()
    command = str(args.command).replace('-','_')

    if args.command is None:
        parser.print_help()
        raise SystemExit('Exiting without command')

    run_kwargs = {}
    if args.k2_prime is not None:
        run_kwargs['k2_prime'] = args.k2_prime

    method = args.method_name.replace('-','_')

    if command=='graphical_analysis':
        graphical_analysis = pet_ga.GraphicalAnalysis(input_tac_path=args.input_tac_path,
                                                    roi_tac_path=args.roi_tac_path,
                                                    output_directory=args.output_directory,
                                                    output_filename_prefix=args.output_filename_prefix,
                                                    method=method,
                                                    fit_thresh_in_mins=args.threshold_in_mins)
        graphical_analysis.run_analysis(**run_kwargs)
        graphical_analysis.save_analysis()
    if command=='graphical_analysis_multitac':
        graphical_analysis = pet_ga.MultiTACGraphicalAnalysis(input_tac_path=args.input_tac_path,
                                                              roi_tacs_dir=args.roi_tacs_dir,
                                                              output_directory=args.output_directory,
                                                              output_filename_prefix=args.output_filename_prefix,
                                                              method=method,
                                                              fit_thresh_in_mins=args.threshold_in_mins)
        graphical_analysis(output_as_tsv=True, output_as_json=False, **run_kwargs)

    if args.print:
        for key, val in graphical_analysis.analysis_props.items():
            print(f"{key:<20}:  {val}")


if __name__ == "__main__":
    main()



"""
Regional TAC extraction
"""
from warnings import warn
import os
from collections.abc import Callable
import pathlib
import numpy as np
import ants
import pandas as pd

from .segmentation_tools import combine_regions_as_mask
from ..utils import image_io
from ..utils.scan_timing import ScanTimingInfo
from ..utils.dimension import check_physical_space_for_ants_image_pair
from ..utils.time_activity_curve import TimeActivityCurve
from ..meta.label_maps import LabelMapLoader

def extract_roi_voxel_tacs_from_image_using_mask(input_image: ants.core.ANTsImage,
                                                 mask_image: ants.core.ANTsImage,
                                                 verbose: bool = False) -> np.ndarray:
    """
    Function to extract ROI voxel tacs from an image using a mask image.

    This function returns all the voxel TACs, and unlike
    :func:`extract_mean_roi_tac_from_nifti_using_segmentation` does not calculate the mean over
    all the voxels.

    Args:
        input_image (ants.core.ANTsImage): Input 4D-image from which to extract ROI voxel tacs.
        mask_image (ants.core.ANTsImage): Mask image which determines which voxels to extract.
        verbose (bool, optional): If True, prints information about the shape of extracted voxel
            tacs.

    Returns:
        out_voxels (np.ndarray): Array of voxel TACs of shape (num_voxels, num_frames)

    Raises:
         AssertionError: If input image is not 4D-image.
         AssertionError: If mask image is not in the same physical space as the input image.

    Example:

        .. code-block:: python

            import ants
            import numpy as np

            from petpal.preproc import regional_tac_extraction
            tac_func = regional_tac_extraction.extract_roi_voxel_tacs_from_image_using_mask
            
            # Read images
            pet_img = ants.image_read("/path/to/pet.nii.gz")
            masked_region_img = ants.image_read("/path/to/mask_region.nii.gz")

            # Run ROI extraction and save
            time_series = tac_func(input_image=pet_img, mask_image=masked_region_img).T
            np.savetxt("time_series.tsv", time_series, delimiter='\t')
            
    """
    assert len(input_image.shape) == 4, "Input image must be 4D."
    assert check_physical_space_for_ants_image_pair(input_image, mask_image), (
        "Images must have the same physical dimensions.")

    out_voxels = apply_mask_4d(input_arr=input_image.numpy(),
                               mask_arr=mask_image.numpy(),
                               verbose=verbose)
    return out_voxels


def apply_mask_4d(input_arr: np.ndarray,
                  mask_arr: np.ndarray,
                  verbose: bool = False) -> np.ndarray:
    """
    Function to extract ROI voxel tacs from an array using a mask array.

    This function applies a 3D mask to a 4D image, returning the time series for each voxel in a
    single flattened numpy array.

    Args:
        input_arr (np.ndarray): Input 4D-image from which to extract ROI voxel tacs.
        mask_arr (np.ndarray): Mask image which determines which voxels to extract.
        verbose (bool, optional): If True, prints information about the shape of extracted voxel
            tacs.

    Returns:
        out_voxels (np.ndarray): Time series of each voxel in the mask, as a flattened numpy array.

    Raises:
         AssertionError: If input array is not 4D.
         AssertionError: If input and mask array shapes are mismatched.

    Example:

        .. code-block:: python

            import ants
            import numpy as np

            from petpal.preproc.regional_tac_extraction import apply_mask_4d
            
            # Read images
            pet_img = ants.image_read("/path/to/pet.nii.gz")
            masked_region_img = ants.image_read("/path/to/mask_region.nii.gz")

            # Get underlying arrays
            pet_arr = pet_img.numpy()
            masked_region_arr = masked_region_img.numpy()

            # Run ROI extraction and save
            time_series = apply_mask_4d(input_arr=pet_arr, mask_arr=masked_region_arr).T
            np.savetxt("time_series.tsv", time_series, delimiter='\t')

    """
    assert len(input_arr.shape) == 4, "Input array must be 4D."
    assert input_arr.shape[:3] == mask_arr.shape, (
            "Array must have the same physical dimensions.")

    x_inds, y_inds, z_inds = mask_arr.nonzero()
    out_voxels = input_arr[x_inds, y_inds, z_inds, :]
    if verbose:
        print(f"(ImageOps): Output TACs have shape {out_voxels.shape}")
    return out_voxels


def voxel_average_w_uncertainty(pet_voxels: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Spatially average flattened PET voxels and get the standard deviation as well.

    Takes a 2D M x N numpy array with M voxels and N time frames and returns a tuple of 1D N-length
    numpy arrays with the mean activity and standard deviation for each time frame.

    Args:
        pet_voxels (np.ndarray): 2D M x N array of PET voxels with M voxels and N time frames.
            Typically takes the output of
            :func:`~petpal.preproc.regional_tac_extraction.apply_mask_4d`.

    Returns:
        average_w_uncertainty (tuple[np.ndarray, np.ndarray]): Average and standard deviation of
            PET voxels for each time frame."""
    pet_average = pet_voxels.mean((0))
    pet_uncertainty = pet_voxels.std((0))
    return (pet_average, pet_uncertainty)


def write_tacs(input_image_path: str,
               label_map_path: str,
               segmentation_image_path: str,
               out_tac_dir: str,
               out_tac_prefix: str = '',
               verbose: bool = False):
    """
    Function to write Tissue Activity Curves for each region, given a segmentation,
    4D PET image, and label map. Computes the average of the PET image within each
    region. Writes a JSON for each region with region name, frame start time, and mean 
    value within region.

    Args:
        input_image_path (str): Path to the 4D PET image from which regional TACs will be
            extracted.
        label_map_path (str): Path to the dseg file linking regions to their mappings in the
            segmentation image.
        segmentation_image_path (str): Path to the segmentation image containing ROIs. Must be in
            the same space as input_image.
        out_tac_dir (str): Path to the directory where regional TACs will be written to.
        out_tac_prefix (str): Prefix for output TAC files. Typically the participant ID.
        verbose (bool): Set to True to print processing info. Default False.
    """
    label_map = image_io.read_label_map_tsv(label_map_file=label_map_path)
    regions_abrev = label_map['abbreviation']

    pet_numpy = ants.image_read(input_image_path).numpy()
    seg_numpy = ants.image_read(segmentation_image_path).numpy()

    scan_timing_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
    tac_times_in_mins = scan_timing_info.center_in_mins

    for i, region_map in enumerate(label_map['mapping']):
        region_mask = combine_regions_as_mask(segmentation_img=seg_numpy,
                                              label=int(region_map))
        pet_masked_region = apply_mask_4d(input_arr=pet_numpy,
                                          mask_arr=region_mask)
        extracted_tac, tac_uncertainty = voxel_average_w_uncertainty(pet_masked_region)
        region_tac = TimeActivityCurve(times=tac_times_in_mins,
                                       activity=extracted_tac,
                                       uncertainty=tac_uncertainty)
        if out_tac_prefix:
            out_tac_path = os.path.join(out_tac_dir,
                                        f'{out_tac_prefix}_seg-{regions_abrev[i]}_tac.tsv')
        else:
            out_tac_path = os.path.join(out_tac_dir, f'seg-{regions_abrev[i]}_tac.tsv')
        region_tac.to_tsv(filename=out_tac_path)
    if verbose:
        print('Finished writing TACs.')


def roi_tac(input_image_path: str,
            roi_image_path: str,
            region: list[int] | int,
            out_tac_path: str | None = None,
            time_frame_keyword: str = 'FrameReferenceTime') -> TimeActivityCurve:
    """
    Function to write Tissue Activity Curves for a single region, given a mask,
    4D PET image, and region mapping. Computes the average of the PET image 
    within each region. Writes a tsv table with region name, frame start time,
    and mean value within region.

    Args:
        input_image_path (str): Path to the 4D PET image from which the regional TAC will be
            extracted.
        roi_image_path (str): Path to the segmentation image containing ROIs. Must be in the same
            space as input_image.
        region (list[int] | int): The region or regions that will be extracted as a TAC. If a list
            of regions are provided, the function combines all listed regions and calculates the
            TAC from the merged region.
        out_tac_dir (str): Path to the TSV where the regional TAC will be written to.
        time_frame_keyword (str): Keyword corresponding to either 'FrameReferenceTime' or
            'FrameTimesStart' to get the frame timing. Default 'FrameReferenceTime'.

    Returns:
        region_tac (TimeActivityCurve): The mean time activity curve for the region.
    """

    if time_frame_keyword not in ['FrameReferenceTime', 'FrameTimesStart']:
        raise ValueError("'time_frame_keyword' must be one of "
                         "'FrameReferenceTime' or 'FrameTimesStart'")

    pet_meta = image_io.load_metadata_for_nifti_with_same_filename(input_image_path)
    pet_numpy = ants.image_read(input_image_path).numpy()
    seg_numpy = ants.image_read(roi_image_path).numpy()

    region_mask = combine_regions_as_mask(segmentation_img=seg_numpy,
                                          label=region)
    pet_masked_region = apply_mask_4d(input_arr=pet_numpy,
                                      mask_arr=region_mask)
    extracted_tac, tac_uncertainty = voxel_average_w_uncertainty(pet_masked_region)
    region_tac = TimeActivityCurve(times=pet_meta[time_frame_keyword],
                                   activity=extracted_tac,
                                   uncertainty=tac_uncertainty)
    if out_tac_path is not None:
        region_tac.to_tsv(filename=out_tac_path)

    return region_tac

class WriteRegionalTacs:
    """
    Write regional TACs

    Attributes:
        pet_arr (np.ndarray): Numpy array containing 4D PET data.
        seg_arr (np.ndarray): Numpy array containing 3D discrete segmentation data.
        tac_extraction_func (Callable): A function that takes a 2D M x N numpy array with M voxels
            and N time frames as well as any number of optional keyword arguments and returns a
            tuple of 1D N-length numpy arrays with the calculated TAC and uncertainty.
        scan_timing (ScanTimingInfo): Scan timing for the input PET image.
        region_names (list): Names of regions to use in the analysis.
        region_maps (list): Region mappings to use in the analysis, corresponding 1-1 with
            region_names.


    Example:

        .. code-block:: python

            from petpal.preproc.regional_tac_extraction import WriteRegionalTacs
            from petpal.utils.bids_utils import gen_bids_like_filepath, gen_bids_like_dir_path
            
            pet_image_path = gen_bids_like_filepath(sub_id='001',
                                                    ses_id='01',
                                                    suffix='pet',
                                                    ext='.nii.gz')
            seg_image_path = gen_bids_like_filepath(sub_id='001',
                                                    ses_id='01',
                                                    bids_dir='../derivatives/petpal/'
                                                    suffix='seg',
                                                    modality='seg',
                                                    space='pet',
                                                    ext='.nii.gz')
            tac_output_dir = gen_bids_like_dir_path(sub_id='001',
                                                    ses_id='01',
                                                    modality='tacs',
                                                    sup_dir='../derivatives/petpal/')
            tac_calculator = WriteRegionalTacs(input_image_path=pet_image_path,
                                               segmentation_path=seg_image_path,
                                               label_map_path='dseg.tsv')
            tac_calculator(out_tac_dir=tac_output_dir,
                           out_tac_prefix='sub-001_ses-01',
                           one_tsv_per_region=False)

    """
    def __init__(self,
                 input_image_path: str | pathlib.Path,
                 segmentation_path: str | pathlib.Path,
                 label_map: str | dict,
                 tac_extraction_func: Callable=voxel_average_w_uncertainty):
        """Initialize WriteRegionalTacs.
        
        Args:
            input_image_path (str | pathlib.Path): Path to input 4D PET image.
            segmentation_path (str | pathlib.Path): Path to 3D discrete segmentation image. Must
                match input PET image space.
            label_map (str | dict): Label map for use in the study. Provide name of a preset
                label map option such as 'freesurfer', the path to a label map JSON file, or a
                Python dictionary with region mappings. For more details, see
                :class:`LabelMapLoader<petpal.meta.label_maps.LabelMapLoader>`.
            tac_extraction_func (Callable): Function to get TAC from 2D array of voxels. Default
                :func:`~petpal.preproc.regional_tac_extraction.voxel_average_w_uncertainty`.
        """
        self.pet_arr = ants.image_read(filename=input_image_path).numpy()
        self.seg_arr = ants.image_read(filename=segmentation_path).numpy()

        self.tac_extraction_func = tac_extraction_func
        self.scan_timing = ScanTimingInfo.from_nifti(input_image_path)

        label_map_dict = LabelMapLoader(label_map_option=label_map).label_map
        self.region_names = list(label_map_dict.keys())
        self.region_maps = list(label_map_dict.values())

    def set_tac_extraction_func(self, tac_extraction_func: Callable):
        """Sets the tac extraction function used to a different function.
        
        The selected function must take a 2D array of the masked voxels as input, and return the 
        calculated activity and uncertainty across the masked voxels outputs.

        Args:
            tac_extraction_func (Callable): Function that takes a 2D M x N numpy array with M
                voxels and N time frames as well as any number of optional keyword arguments and
                returns a tuple of 1D N-length numpy arrays with the calculated TAC and
                uncertainty.
        """
        self.tac_extraction_func = tac_extraction_func


    def find_label_name(self, label: int) -> str:
        """Find the name for a label based on the provided label map. If a name is not found,
        return 'UNK' followed by the label index.
        
        Args:
            label (int): Label mapping for a region.
        
        Returns:
            region_name (str): Name of the region corresponding to the provided label."""
        try:
            label_map_loc = self.region_maps.index(label)
            region_name = self.region_names[label_map_loc]
        except ValueError:
            region_name = f'UNK{label:>04}'
        return region_name

    def is_empty_region(self,pet_masked_region: np.ndarray) -> bool:
        """Check if masked PET region has zero matched voxels, or is all NaNs. In either case,
        return True, otherwise return False.
        
        Args:
            pet_masked_region (np.ndarray): Array of PET voxels masked to a specific region.
            
        Returns:
            pet_masked_region_is_empty (bool): If True, input region is empty."""
        if pet_masked_region.size==0:
            return True
        if np.all(np.isnan(pet_masked_region)):
            return True
        return False

    def extract_tac(self,region_mapping: int | list[int], **tac_calc_kwargs) -> TimeActivityCurve:
        """
        Run self.tac_extraction_func on one region and return the TAC.

        Args:
            region_mapping (int | list[int]): The integer ID or IDs corresponding to the ROI.
            **tac_calc_kwargs: Additional keyword arguments passed on to tac_extraction_func.
    
        Returns:
            region_tac (TimeActivityCurve): The calculated TAC for the region. 
        """
        region_mask = combine_regions_as_mask(segmentation_img=self.seg_arr,
                                              label=region_mapping)

        pet_masked_region = apply_mask_4d(input_arr=self.pet_arr,
                                          mask_arr=region_mask)

        is_region_empty = self.is_empty_region(pet_masked_region=pet_masked_region)
        if is_region_empty:
            extracted_tac = np.empty_like(self.scan_timing.center_in_mins)
            extracted_tac.fill(np.nan)
            uncertainty = extracted_tac.copy()
        else:
            extracted_tac, uncertainty = self.tac_extraction_func(pet_voxels=pet_masked_region,
                                                                  **tac_calc_kwargs)
        region_tac = TimeActivityCurve(times=self.scan_timing.center_in_mins,
                                       activity=extracted_tac,
                                       uncertainty=uncertainty)
        return region_tac

    def gen_tacs_data_frame(self) -> pd.DataFrame:
        """Get empty data frame to store TACs. Sets first two columns to frame start and end
        times, and remaining columns are named by region activity and uncertainty, based on the
        regions included in the label map.
        
        Returns:
            tacs_data (pd.DataFrame): Data frame with columns set for frame start and end time,
            and activity and uncertainty for each included region. Frame start and end time
            columns filled with scan timing data.
        """
        activity_uncertainty_column_names = []
        for region_name in self.region_names:
            activity_uncertainty_column_names.append(region_name)
            activity_uncertainty_column_names.append(f'{region_name}_unc')
        cols_list = ['frame_start(min)','frame_end(min)'] + activity_uncertainty_column_names
        tacs_data = pd.DataFrame(columns=cols_list)

        tacs_data['frame_start(min)'] = self.scan_timing.start_in_mins
        tacs_data['frame_end(min)'] = self.scan_timing.end_in_mins

        return tacs_data



    def write_tacs(self,
                   out_tac_prefix: str,
                   out_tac_dir: str | pathlib.Path,
                   one_tsv_per_region: bool=True,
                   **tac_calc_kwargs):
        """
        Function to write Tissue Activity Curves for each region, given a segmentation,
        4D PET image, and label map. Computes the average of the PET image within each
        region. Writes TACs in TSV format with region name, frame start time, frame end time, and
        activity and uncertainty within each region. Skips writing regions without any matched
        voxels.

        Args:
            out_tac_prefix (str): Prefix for the output files, usually the BIDS subject and
                session ID.
            out_tac_dir (str | pathlib.Path): Output path where files are saved.
            one_tsv_per_region (bool): If True, write one TSV TAC file for each region in the
                image. If False, write one TSV file with all TACs in the image.
            **tac_calc_kwargs: Additional keywords passed onto tac_extraction_func.

        Raises:
            Warning: for each region without any matched voxels, warn user that TAC is skipped.
        """
        tacs_data = self.gen_tacs_data_frame()

        empty_regions = []
        for i,region_name in enumerate(self.region_names):
            mappings = self.region_maps[i]
            tac = self.extract_tac(region_mapping=mappings, **tac_calc_kwargs)
            if tac.contains_any_nan:
                empty_regions.append(region_name)
                continue
            if one_tsv_per_region:
                os.makedirs(out_tac_dir, exist_ok=True)
                tac.to_tsv(filename=f'{out_tac_dir}/{out_tac_prefix}_seg-{region_name}_tac.tsv')
            else:
                tacs_data[region_name] = tac.activity
                tacs_data[f'{region_name}_unc'] = tac.uncertainty

        if len(empty_regions)>0:
            warn("Empty regions were found during tac extraction. TACs for the following regions "
                 f"were not saved: {empty_regions}")
            tacs_data.drop(empty_regions,axis=1,inplace=True)
            tacs_data.drop([f'{region}_unc' for region in empty_regions],axis=1,inplace=True)

        if not one_tsv_per_region:
            os.makedirs(out_tac_dir, exist_ok=True)
            tacs_data.to_csv(f'{out_tac_dir}/{out_tac_prefix}_multitacs.tsv', sep='\t', index=False)

    def __call__(self,
                 out_tac_prefix: str,
                 out_tac_dir: str | pathlib.Path,
                 one_tsv_per_region: bool=True,
                 **tac_calc_kwargs):
        """Runs TAC computation and writing by running `self.write_tacs`.
        
        Args:
            out_tac_prefix (str): Prefix for the output files, usually the BIDS subject and
                session ID.
            out_tac_dir (str | pathlib.Path): Output path where files are saved.
            one_tsv_per_region (bool): If True, write one TSV TAC file for each region in the
                image. If False, write one TSV file with all TACs in the image.
            **tac_calc_kwargs: Additional keywords passed onto tac_extraction_func."""
        self.write_tacs(out_tac_prefix=out_tac_prefix,
                        out_tac_dir=out_tac_dir,
                        one_tsv_per_region=one_tsv_per_region,
                        **tac_calc_kwargs)



"""
Data driven image analyses, including PCA analysis and input function extraction.
"""
import ants
import numpy as np
from sklearn.decomposition import PCA

from ..preproc.regional_tac_extraction import extract_roi_voxel_tacs_from_image_using_mask
from .dimension import check_physical_space_for_ants_image_pair
from .scan_timing import ScanTimingInfo


def temporal_pca_analysis_of_image_over_mask(input_image: ants.core.ANTsImage,
                                             mask_image: ants.core.ANTsImage,
                                             num_components: int = 3,
                                             svd_solver: str = 'full',
                                             whiten: bool = True,
                                             **sklearn_pca_kwargs) -> tuple[PCA, np.ndarray]:
    """
    Perform temporal PCA (Principal Component Analysis) on a 4D PET image using a specified mask.

    This function analyzes the temporal dynamics of voxel time-activity curves (TACs) extracted
    from the region specified by the input mask. PCA reduces the dimensionality of the dataset,
    extracting a set of temporal features ranked by their contribution to variability across
    voxels within the specified region. This approach effectively identifies the most prominent
    temporal patterns in the PET image's dynamic data.

    .. important::
        Ensure that the input and mask images are in the same coordinate space
        (aligned and registered) to avoid errors or mismatched results.

    Args:
        input_image (ants.core.ANTsImage):
            The 4D PET image for temporal PCA analysis. The image must have 3 spatial dimensions
            and one temporal dimension.
        mask_image (ants.core.ANTsImage):
            A binary mask representing the region of interest (ROI) in the input image.
            Only voxels within this mask are included in the PCA computation. The mask and input
            image must have the same physical dimensions.
        num_components (int, optional):
            The number of principal components to compute. Defaults to 3.
        svd_solver (str, optional): The type of SVD solver to use for PCA. Defaults to `"full"`.
        whiten (bool, optional):
            If True, the output features are normalized to have unit variance. This is useful
            for further statistical analysis of the temporal components. Defaults to True.
        **sklearn_pca_kwargs:
            Additional keyword arguments to customize the PCA behavior using scikit-learn's
            `PCA` class (e.g., `random_state`, `tol`, etc.).

    Returns:
        tuple[PCA, np.ndarray]: Fitted PCA model object and voxel TACs projected onto the principal components.

    Example:

        .. code-block:: python

            import ants
            from petpal.utils.data_driven_image_analyses import temporal_pca_analysis_of_image_over_mask as tpca_func

            pca_obj, pca_projections = tpca_func(input_image=ants.image_read('/path/to/4D/PET.nii.gz'),
                                                 mask_image=ants.image_read('/path/to/aligned/mask.nii.gz'),
                                                 num_components=3)

    Raises:
        AssertionError: The input image is not 4D.
        AssertionError: The input image and mask image are not in the same physical space or dimensions.

    Notes:
        - Ensure that the input image and mask image are pre-registered and aligned, with identical
          spatial dimensions.
        - The PCA algorithm identifies patterns across the temporal dynamics while minimizing loss
          of information.
        - Whitening normalizes the data, which can be important if further statistical operations
          (e.g., clustering) are planned.

    See Also:
        - :func:`extract_roi_voxel_tacs_from_image_using_mask<petpal.preproc.regional_tac_extraction.extract_roi_voxel_tacs_from_image_using_mask>`: Extracts voxel-level TACs for regions of interest
          in the input image, used in TAC calculations here.
        - :class:`sklearn.decomposition.PCA`: Underlying PCA implementation used in this function.
    """
    assert len(input_image.shape) == 4, "Input image must be 4D."
    assert check_physical_space_for_ants_image_pair(input_image, mask_image), (
        "Images must have the same physical dimensions.")
    mask_voxels = extract_roi_voxel_tacs_from_image_using_mask(input_image=input_image,
                                                               mask_image=mask_image)
    pca_obj = PCA(n_components=num_components, svd_solver=svd_solver, whiten=whiten, **sklearn_pca_kwargs)
    pca_obj.fit(mask_voxels)
    return pca_obj, np.ascontiguousarray(pca_obj.fit_transform(mask_voxels), float)


def extract_temporal_pca_components_of_image_over_mask(input_image: ants.core.ANTsImage,
                                                       mask_image: ants.core.ANTsImage,
                                                       num_components: int = 3,
                                                       svd_solver: str = 'full',
                                                       whiten: bool = True,
                                                       **sklearn_pca_kwargs) -> np.ndarray:
    """
    Extract principal components from temporal PCA on a 4D PET image using a specified mask.

    This function performs temporal Principal Component Analysis (PCA) on a 4D input image,
    focusing on the region defined by the mask. It returns only the PCA components
    (i.e., the eigenvectors that describe temporal variation) from the fitted PCA model.
    These components might help you identify the primary kinetic components over the mask.

    .. important::
        Ensure that the input and mask images are in the same coordinate space
        (aligned and registered) to avoid errors or mismatched results.

    Args:
        input_image (ants.core.ANTsImage):
            The 4D PET image for temporal PCA analysis. The image should have 3 spatial
            dimensions and one temporal dimension.
        mask_image (ants.core.ANTsImage):
            A binary mask defining the region of interest (ROI) in the input image.
            Only voxels within this mask will be considered during PCA computation.
            The mask and the input image must share the same physical and spatial dimensions.
        num_components (int, optional):
            The number of principal components to compute. Defaults to 3.
        svd_solver (str, optional):
            The type of SVD solver to use for PCA. Defaults to `"full"`.
        whiten (bool, optional):
            If set to `True`, the resulting components are scaled to have unit variance.
            Defaults to `True`.
        **sklearn_pca_kwargs:
            Additional keyword arguments for customizing PCA behavior via scikit-learn's
            `PCA` class, such as `random_state`, `tol`, etc.

    Returns:
        np.ndarray:
            A 2D array of shape `(num_components, num_timepoints)` representing the PCA
            components of temporal features. Each row corresponds to a principal component
            in the temporal domain.

    Example:

        .. code-block:: python

            import ants
            from petpal.utils.data_driven_image_analyses import extract_temporal_pca_components_of_image_over_mask as ext_tpca_comps_func

            pca_comps = ext_tpca_comps_func(input_image=ants.image_read('/path/to/4D/PET.nii.gz'),
                                            mask_image=ants.image_read('/path/to/aligned/mask.nii.gz'),
                                            num_components=3)


    Raises:
        AssertionError:
            Raised if the input image is not 4D, or if the mask is not in the same
            physical space as the input image.
        ValueError:
            Raised if the number of components exceeds the valid limits given the input
            data size.

    Notes:
        - This function is a higher-level utility that focuses solely on extracting
          PCA components for temporal dynamics.

    See Also:
        - :func:`temporal_pca_analysis_of_image_over_mask`: Performs full temporal PCA
          and returns both the fitted PCA model and the projections of voxel time-activity
          curves.
        - :class:`sklearn.decomposition.PCA`: Core class for performing Principal
          Component Analysis in scikit-learn.

    """
    pca_obj, _ = temporal_pca_analysis_of_image_over_mask(input_image=input_image,
                                                          mask_image=mask_image,
                                                          num_components=num_components,
                                                          svd_solver=svd_solver,
                                                          whiten=whiten,
                                                          **sklearn_pca_kwargs)
    return pca_obj.components_


def extract_temporal_pca_projection_of_image_over_mask(input_image: ants.core.ANTsImage,
                                                       mask_image: ants.core.ANTsImage,
                                                       num_components: int = 3,
                                                       svd_solver: str = 'full',
                                                       whiten: bool = True,
                                                       **sklearn_pca_kwargs) -> np.ndarray:
    """
    Compute and extract the temporal PCA projections for a masked region in a 4D PET image.

    This function carries out temporal Principal Component Analysis (PCA) on a 4D input image,
    focusing on the region defined by the mask. It returns the projections (i.e., the representation
    of each voxel in the reduced-dimensional space of principal components).

    .. important::
        Ensure that the input and mask images are in the same coordinate space
        (aligned and registered) to avoid errors or mismatched results.

    Args:
        input_image (ants.core.ANTsImage):
            A 4D PET image for temporal PCA analysis. The image should have 3 spatial
            dimensions and one temporal dimension.
        mask_image (ants.core.ANTsImage):
            A binary mask specifying the region of interest (ROI) to include in the PCA analysis.
            The mask must have the same physical and spatial dimensions as the input image.
        num_components (int, optional):
            The number of principal components to compute. Determines the dimensionality of
            the reduced output space. Defaults to 3.
        svd_solver (str, optional):
            Determines the solver used for Singular Value Decomposition (SVD) during PCA.
            Defaults to `"full"`.
        whiten (bool, optional):
            If set to `True`, components are scaled to have unit variance. This is often
            useful when the output is used as input to further statistical or machine learning models.
            Defaults to `True`.
        **sklearn_pca_kwargs:
            Additional keyword arguments for scikit-learn's `PCA` class, such as specification
            of `random_state`, `tol`, etc.

    Returns:
        np.ndarray:
            A 2D array of shape `(num_voxels, num_components)` containing the temporal PCA
            projections for each voxel within the mask. Each row represents the reduced-dimensional
            representation of a voxel's TAC within the selected components.

    Example:

        .. code-block:: python

            import ants
            from petpal.utils.data_driven_image_analyses import extract_temporal_pca_projection_of_image_over_mask as ext_tpca_proj_func

            pca_proj = ext_tpca_proj_func(input_image=ants.image_read('/path/to/4D/PET.nii.gz'),
                                          mask_image=ants.image_read('/path/to/aligned/mask.nii.gz'),
                                          num_components=3)

    Raises:
        AssertionError:
            Raised if the `input_image` is not 4D, or if the physical dimensions of the
            `mask_image` and `input_image` are not identical.
        ValueError:
            Raised if the number of PCA components exceeds the number of voxels included
            in the mask.

    Notes:
        - The projections represent the linear combination of the temporal principal components
          for each voxel, condensed into a reduced space.

    See Also:
        - :func:`temporal_pca_analysis_of_image_over_mask`: Performs the underlying PCA
          calculation and provides both PCA projections and the PCA model.
        - :class:`sklearn.decomposition.PCA`: Core implementation of the PCA used in this function.
    """
    _, transformed_voxels = temporal_pca_analysis_of_image_over_mask(input_image=input_image,
                                                                     mask_image=mask_image,
                                                                     num_components=num_components,
                                                                     svd_solver=svd_solver,
                                                                     whiten=whiten,
                                                                     **sklearn_pca_kwargs)
    return transformed_voxels


def extract_temporal_pca_quantile_thresholded_tacs_of_image_using_mask(input_image: ants.core.ANTsImage,
                                                                       mask_image: ants.core.ANTsImage,
                                                                       num_components: int = 3,
                                                                       threshold_components: list[int] | None = None,
                                                                       quantiles: list[float] | None = None,
                                                                       direction: str = '>',
                                                                       **sklearn_pca_kwargs) -> np.ndarray:
    """
    Extract quantile-thresholded time-activity curve (TAC) values for temporal PCA components from a 4D PET image.

    This function computes time-activity curve (TAC) statistics for masked regions in a 4D PET image, based on
    Principal Component Analysis (PCA). It thresholds the PCA projections (for provided selected components) using
    specified quantiles, and calculates the mean and standard deviation of the TACs for voxels exceeding the
    thresholds. The output includes both the TAC mean and standard deviation across all components and quantile
    thresholds.

    .. important::
        Ensure that the input and mask images are in the same coordinate space
        (aligned and registered) to avoid errors or mismatched results.

    Args:
        input_image (ants.core.ANTsImage):
            A 4D PET image.
        mask_image (ants.core.ANTsImage):
            A binary mask specifying the region of interest (ROI) in the image. Only voxels within this mask
            are considered for the PCA analysis and TAC computation. This image is assumed to be aligned
            to the input image.
        num_components (int, optional):
            The number of PCA components to use for reducing the dataset's dimensionality. Defaults to 3.
        threshold_components (list[int] | None, optional):
            A list of component indices (from 0 to `num_components - 1`) to use for thresholding.
            Thresholding is applied to these PCA components using the given quantiles. If not set, defaults to `[0]`.
        quantiles (list[float], optional):
            An array of quantiles (values between 0 and 1) to compute thresholds for the PCA projections.
            Only voxels with PCA projection values above the corresponding quantile thresholds are included
            when calculating TAC statistics. Defaults to `[0.5, 0.75, 0.9, 0.975]`.
        direction (str, optional):
            An optional argument to switch from ``>`` to ``<``. Defaults to ``>``.
        **sklearn_pca_kwargs:
            Additional keyword arguments passed to customize scikit-learn's PCA class (e.g., `svd_solver`,
            `random_state`, etc.).

    Returns:
        np.ndarray:
            A 2-element array containing `out_vals` (*np.ndarray*),  a 3D array of shape
            `(num_components, num_quantiles, num_timepoints)`, containing the mean TAC values of
            voxels exceeding the quantile thresholds for each component, and `out_stds` (*np.ndarray*),
            a 3D array of the same shape as `out_vals`, containing the standard deviation of the TAC
            values under the same conditions.

    Example:

        .. code-block:: python

            import ants
            from petpal.utils.data_driven_image_analyses import extract_temporal_pca_quantile_thresholded_tacs_of_image_using_mask as ext_pca_q_func

            tac_means, tac_stds = ext_pca_q_func(input_image=ants.image_read('/path/to/4D/PET.nii.gz'),
                                                 mask_image=ants.image_read('/path/to/aligned/mask.nii.gz'),
                                                 num_components=3,
                                                 threshold_components=[0,1,2],
                                                 direction='>', # set to '<' if we want the opposite direction
                                                 quantiles=np.array([0.5, 0.75, 0.9, 0.975]))

    Raises:
        AssertionError: `threshold_components` contains invalid indices (e.g., negative or greater than `num_components - 1`).
        AssertionError: Quantiles are not within the [0, 1] range.

    Notes:
        - Thresholding PCA projections using quantiles isolates specific temporal dynamics for analysis,
          which can provide insight into dominant trends in the time-activity data.
        - Voxels not exceeding the quantile thresholds are excluded from the mean and standard deviation
          calculations.

    See Also:
        - :func:`extract_temporal_pca_projection_of_image_over_mask`: Computes PCA projections used as input
          for thresholding in this function.
        - :func:`extract_roi_voxel_tacs_from_image_using_mask<petpal.preproc.regional_tac_extraction.extract_roi_voxel_tacs_from_image_using_mask>`: Extracts voxel-level TACs for regions of interest
          in the input image, used in TAC calculations here.
        - :class:`sklearn.decomposition.PCA`: Core PCA implementation used in this analysis.
    """

    if threshold_components is None:
        threshold_components = [i for i in range(num_components)]

    if quantiles is None:
        quantiles = [0.5, 0.75, 0.9, 0.975]

    assert min(threshold_components) >= 0, "Threshold components must be integers >= 0."
    assert max(threshold_components) < num_components, "Threshold components must be < num_components."
    assert np.min(quantiles) >= 0, "Quantiles must be >= 0."
    assert np.max(quantiles) <= 1, "Quantiles must be <= 1."

    voxels_pca_projs = extract_temporal_pca_projection_of_image_over_mask(input_image=input_image,
                                                                          mask_image=mask_image,
                                                                          num_components=num_components,
                                                                          **sklearn_pca_kwargs)
    voxels_pca_projs = voxels_pca_projs[:, threshold_components]

    roi_voxels = extract_roi_voxel_tacs_from_image_using_mask(input_image=input_image, mask_image=mask_image)

    num_components = len(threshold_components)
    num_quantiles = len(quantiles)
    num_frames = roi_voxels.shape[1]

    tacs_mean = np.zeros((num_components, num_quantiles, num_frames), float)
    tacs_std = np.zeros_like(tacs_mean)


    for comp_id, comp_pc_vals in enumerate(voxels_pca_projs.T):
        thresh_vals_for_comp = np.quantile(comp_pc_vals, quantiles)
        for thresh_id, a_thresh in enumerate(thresh_vals_for_comp):
            valid_voxels_mask = comp_pc_vals < a_thresh if direction == '<' else comp_pc_vals > a_thresh
            tmp_tacs = roi_voxels[valid_voxels_mask, :]
            tacs_mean[comp_id, thresh_id, :] = np.nanmean(tmp_tacs, axis=0)
            tacs_std[comp_id, thresh_id, :] = np.nanstd(tmp_tacs, axis=0)

    return np.asarray([tacs_mean, tacs_std])





def generate_temporal_pca_quantile_threshold_tacs_of_image_over_mask(input_image_path: str,
                                                                     mask_image_path: str,
                                                                     output_arrays_path: str | None,
                                                                     num_components: int,
                                                                     threshold_components: list | None,
                                                                     quantiles: list | None,
                                                                     direction: str = '>',
                                                                     **sklearn_pca_kwargs
                                                                     ):
    """
    Extract and compute time-activity curves (TACs) for an image using components of PCA projections
    and quantile thresholds over a specified mask.

    .. important::
        Ensure that the input and mask images are in the same coordinate space
        (aligned and registered) to avoid errors or mismatched results.

    This function processes a 4D-PET image (`input_image`) with an accompanying mask,
    computes PCA projections on the image data within the mask, thresholds these projections 
    using specified quantiles, and extracts TAC statistics (mean and standard deviation) 
    for each component-quantile pair.

    The results can optionally be saved to a file, and the TAC means and standard deviations 
    are returned as a NumPy array. The file written to disk has a different format than the array output
    of this function. The written file will have the frame reference times as the first column. The rest of
    the columns will come in pairs of mean and stderr of the TACs, iterated over the quantiles and then
    the threshold components.

    Args:
        input_image_path (str): Path to the input image file (in NIfTI format or similar).
        mask_image_path (str): Path to the mask image file (same dimensions as `input_image`).
        output_arrays_path (str | None):
            Path to save the output TAC arrays. If `None`, the arrays are not saved.
        num_components (int): Number of PCA components to compute.
        threshold_components (list[int] | None):
            Indices of PCA components to apply thresholds on. Defaults to `[0, 1]` if `None`.
        quantiles (list[float] | None):
            Array of quantile thresholds for filtering the PCA projections.
            Defaults to `[0.5, 0.75, 0.9, 0.975]` if `None`.
        direction (str):
            Comparison direction for the thresholds. Defaults to '>' (projection values
            greater than the quantile threshold). Other options may include '<'.
        **sklearn_pca_kwargs:
            Additional keyword arguments to pass to the PCA implementation (from scikit-learn), if applicable.

    Returns:
        np.ndarray:
            A NumPy array containing the mean and standard deviation TACs for each component-quantile pair,
            with shape `(2, num_components, num_quantiles, num_frames)`.

    Example:

        .. code-block:: python

                from petpal.utils.data_driven_image_analyses import generate_temporal_pca_quantile_threshold_tacs_of_image_over_mask


                # Input paths
                input_image_path = "example_input_image.nii"
                mask_image_path = "example_mask_image.nii"
                output_arrays_path = "output_tacs.txt"

                # Parameters
                num_components = 3
                threshold_components = [0, 1]
                quantiles = [0.5, 0.75, 0.95]
                direction = '<'

                # Run function
                tpca_tacs_func = generate_temporal_pca_quantile_threshold_tacs_of_image_over_mask
                tac_means, tac_stds = tpca_tacs_func(input_image_path=input_image_path,
                                                     mask_image_path=mask_image_path,
                                                     output_arrays_path=output_arrays_path,
                                                     num_components=num_components,
                                                     threshold_components=threshold_components,
                                                     quantiles=quantiles,
                                                     direction=direction)
                # Output shape
                print(tac_means.shape)

    Raises:
        AssertionError: If the shapes of the computed TAC mean and standard deviation arrays do not match.

    See Also:
        * :class:`sklearn.decomposition.PCA`: Core PCA implementation used in this analysis.
        * :func:`extract_temporal_pca_quantile_thresholded_tacs_of_image_using_mask`: computes PCA projections
        * :func:`_generate_quantiled_multi_tacs_header`: formats the output array header
        * :func:`_gen_reshaped_quantiled_tacs`: reshapes the means and stds for writing to disk

    """
    if threshold_components is None:
        threshold_components = [0, 1]

    if quantiles is None:
        quantiles = [0.5, 0.75, 0.9, 0.975]

    input_image = ants.image_read(input_image_path)
    mask_image = ants.image_read(mask_image_path)

    tacs_ext_func = extract_temporal_pca_quantile_thresholded_tacs_of_image_using_mask
    tacs_mean, tacs_std = tacs_ext_func(input_image=input_image,
                                        mask_image=mask_image,
                                        num_components=num_components,
                                        threshold_components=threshold_components,
                                        quantiles=quantiles,
                                        direction=direction,
                                        **sklearn_pca_kwargs
                                        )

    if output_arrays_path is not None:
        image_timing_info = ScanTimingInfo.from_nifti(image_path=input_image_path)
        tac_ref_times = image_timing_info.center_in_mins
        out_array = _gen_reshaped_quantiled_tacs(times=tac_ref_times,
                                                 tacs_mean=tacs_mean,
                                                 tacs_std=tacs_std)
        out_header = _generate_quantiled_multi_tacs_header(threshold_components=threshold_components,
                                                           quantiles=quantiles,
                                                           direction=direction)
        np.savetxt(fname=output_arrays_path,
                   X=out_array.T,
                   fmt='%.6e',
                   delimiter='\t',
                   header=out_header,
                   comments='')

    return np.asarray([tacs_mean, tacs_std])


def _generate_quantiled_multi_tacs_header(threshold_components: list, quantiles: list, direction: str):
    """
    Generate the header for outputting multiple TACs generated by the quantile threshold function.

    First column is assumed to be time in minutes, the second column is the mean tac values
    of the first component and first quantile, the third column is the standard deviation,
    and the rest of pairs of columns go over all the quantiles and then components in order.

    Args:
        threshold_components (list[int]): Indices of PCA components for thresholding.
        quantiles (list[float]): List of quantile threshold values.
        direction (str): '<' or '>'.

    Returns:
        str: A single string representing the tab-delimited header.
    """
    header = ['time(mins)']

    for a_comp in threshold_components:
        for a_quan in quantiles:
            header.append(f'(c:{a_comp};q{direction}{a_quan:.3f})_mean')
            header.append(f'(c:{a_comp};q{direction}{a_quan:.3f})_std')

    return "\t".join(header)


def _gen_reshaped_quantiled_tacs(times: np.ndarray[float],
                                 tacs_mean: np.ndarray[float],
                                 tacs_std: np.ndarray[float]) -> np.ndarray:
    """
    Generate the aggregated multiple-tacs array for quantile threshold obtained tacs.

    First column is assumed to be time in minutes, the second column is the mean tac values
    of the first component and first quantile, the third column is the standard deviation,
    and the rest of pairs of columns go over all the quantiles and then components in order.

    Args:
        times (np.ndarray[float]): The frame times corresponding to the TACs.
        tacs_mean (np.ndarray[float]):
            The mean tac values of the TACs over the threshold components and quantiles.
        tacs_std (np.ndarray[float]):
            The stderr value of the TACs over the threshold components and quantiles.
    Returns:
        np.ndarray:
            Array containing the aggregated multiple-tacs array. The first column is the time,
            the rest of the column pairs correspond to the threshold component and the quantile.
    """
    assert tacs_mean.shape == tacs_std.shape, (f"The TACs mean and stderrs must have the same shape."
                                               f" mean:{tacs_mean.shape}, std:{tacs_std.shape}")
    num_components, num_quantiles, num_frames = tacs_mean.shape
    num_tacs = num_components * num_quantiles
    num_columns = 2 * num_tacs + 1
    out_tacs_array = np.zeros(shape=(num_columns, num_frames), dtype=float)

    out_tacs_array[0] = times
    out_tacs_array[1::2] = tacs_mean.reshape(num_tacs, num_frames)
    out_tacs_array[2::2] = tacs_std.reshape(num_tacs, num_frames)

    return out_tacs_array



import warnings
import copy
from typing import Union

from ..preproc.regional_tac_extraction import write_tacs

from .steps_base import *
from ..preproc.image_operations_4d import SimpleAutoImageCropper, rescale_image
from ..preproc.register import register_pet, warp_pet_to_atlas
from ..preproc.motion_corr import windowed_motion_corr_to_target
from ..input_function import blood_input
from ..utils.bids_utils import parse_path_to_get_subject_and_session_id, snake_to_camel_case, gen_bids_like_dir_path, gen_bids_like_filepath
from ..utils.image_io import safe_copy_meta
from ..utils.decorators import ANTsImageToANTsImage

class TACsFromSegmentationStep(FunctionBasedStep):
    """
    A step in a processing pipeline for generating Time Activity Curves (TACs) from segmented images.

    This class is specialized for handling the input and output paths related to TAC generation,
    extending the :class:`FunctionBasedStep<petpal.pipelines.steps_base.FunctionBasedStep>` with specific properties
    and methods for TACs. The class uses :func:`write_tacs<petpal.preproc.regional_tac_extraction.write_tacs>` which
    uses segmentation information to generate ROI TACs, and write them to disk.

    Attributes:
        input_image_path (str): Path to the input image.
        segmentation_image_path (str): Path to the segmentation image.
        segmentation_label_map_path (str): Path to the segmentation label map.
        out_tacs_dir (str): Directory where the output TACs will be saved.
        out_tacs_prefix (str): Prefix for the output TACs.
        verbose (bool): Verbosity flag, default is False.

    """
    def __init__(self,
                 input_image_path: str,
                 segmentation_image_path: str,
                 segmentation_label_map_path: str,
                 out_tacs_dir: str,
                 out_tacs_prefix: str,
                 verbose=False) -> None:
        """
        Initializes a TACsFromSegmentationStep with specified parameters.

        Args:
            input_image_path (str): Path to the input image.
            segmentation_image_path (str): Path to the segmentation image.
            segmentation_label_map_path (str): Path to the segmentation label map.
            out_tacs_dir (str): Directory where the output TACs will be saved.
            out_tacs_prefix (str): Prefix for the output TACs.
            verbose (bool): Verbosity flag, default is False.
        """
        super().__init__(name='write_roi_tacs', function=write_tacs, input_image_path=input_image_path,
                         segmentation_image_path=segmentation_image_path, label_map_path=segmentation_label_map_path,
                         out_tac_dir=out_tacs_dir, out_tac_prefix=out_tacs_prefix,
                         verbose=verbose, )
        self._input_image = input_image_path
        self._segmentation_image = segmentation_image_path
        self._segmentation_label_map = segmentation_label_map_path
        self._out_tacs_path = out_tacs_dir
        self._out_tacs_prefix = out_tacs_prefix
        self.verbose = verbose
    
    def __repr__(self):
        """
        Provides an unambiguous string representation of the TACsFromSegmentationStep instance.

        Returns:
            str: A string representation showing how the instance can be recreated.
        """
        cls_name = type(self).__name__
        info_str = [f'{cls_name}(']
        
        in_kwargs = ArgsDict(
            dict(input_image_path=self.input_image_path, segmentation_image_path=self.segmentation_image_path,
                 segmentation_label_map_path=self.segmentation_label_map_path, out_tacs_dir=self.out_tacs_dir,
                 out_tacs_prefix=self.out_tacs_prefix, verbose=self.verbose))
        
        for arg_name, arg_val in in_kwargs.items():
            info_str.append(f'{arg_name}={repr(arg_val)},')
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    @property
    def segmentation_image_path(self):
        """
        Gets the path to the segmentation image.

        Returns:
            str: The path to the segmentation image.
        """
        return self._segmentation_image
    
    @segmentation_image_path.setter
    def segmentation_image_path(self, segmentation_image_path: str):
        """
        Sets the path to the segmentation image and updates the function arguments.

        Args:
            segmentation_image_path (str): The new path to the segmentation image.
        """
        self._segmentation_image = segmentation_image_path
        self.kwargs['segmentation_image_path'] = segmentation_image_path
    
    @property
    def segmentation_label_map_path(self):
        """
        Gets the path to the segmentation label map.

        Returns:
            str: The path to the segmentation label map.
        """
        return self._segmentation_label_map
    
    @segmentation_label_map_path.setter
    def segmentation_label_map_path(self, segmentation_label_map_path: str):
        """
        Sets the path to the segmentation label map and updates the function arguments.

        Args:
            segmentation_label_map_path (str): The new path to the segmentation label map.
        """
        self._segmentation_label_map = segmentation_label_map_path
        self.kwargs['label_map_path'] = segmentation_label_map_path
    
    @property
    def out_tacs_dir(self):
        """
        Gets the directory where the output TACs will be saved.

        Returns:
            str: The output directory path.
        """
        return self._out_tacs_path
    
    @out_tacs_dir.setter
    def out_tacs_dir(self, out_tacs_path: str):
        """
        Sets the directory where the output TACs will be saved and updates the function arguments.

        Args:
            out_tacs_path (str): The new output directory path.
        """
        self.kwargs['out_tac_dir'] = out_tacs_path
        self._out_tacs_path = out_tacs_path
    
    @property
    def out_tacs_prefix(self):
        """
        Gets the prefix for the output TACs.

        Returns:
            str: The prefix for the output TACs.
        """
        return self._out_tacs_prefix
    
    @out_tacs_prefix.setter
    def out_tacs_prefix(self, out_tacs_prefix: str):
        """
        Sets the prefix for the output TACs and updates the function arguments.

        Args:
            out_tacs_prefix (str): The new prefix for the output TACs.
        """
        self.kwargs['out_tac_prefix'] = out_tacs_prefix
        self._out_tacs_prefix = out_tacs_prefix
    
    @property
    def out_path_and_prefix(self):
        """
        Gets the output directory path and prefix as a tuple.

        Returns:
            tuple: A tuple containing the output directory path and prefix.
        """
        return self._out_tacs_path, self._out_tacs_prefix
    
    @out_path_and_prefix.setter
    def out_path_and_prefix(self, out_dir_and_prefix: str):
        """
        Sets both the output directory path and prefix.

        Args:
            out_dir_and_prefix (tuple): A tuple containing the output directory and prefix.

        Raises:
            ValueError: If the provided value is not a tuple with two items.
        """
        try:
            out_dir, out_prefix = out_dir_and_prefix
        except ValueError:
            raise ValueError("Pass a tuple with two items: `(out_dir, out_prefix)`")
        else:
            self.out_tacs_dir = out_dir
            self.out_tacs_prefix = out_prefix
    
    @property
    def input_image_path(self):
        """
        Gets the path to the input image.

        Returns:
            str: The path to the input image.
        """
        return self._input_image
    
    @input_image_path.setter
    def input_image_path(self, input_image_path: str):
        """
        Sets the path to the input image and updates the function arguments.

        Args:
            input_image_path (str): The new path to the input image.
        """
        self.kwargs['input_image_path'] = input_image_path
        self._input_image = input_image_path
    
    def set_input_as_output_from(self, *sending_steps):
        """
        Sets the input image path based on the output from a specified sending step.

        Args:
            sending_step: The step from which to derive the input image path.
        """
        assert len(sending_steps) == 1, "TACsFromSegmentationStep can only receive a single sending step."
        for sending_step in sending_steps:
            if isinstance(sending_step, ImageToImageStep):
                self.input_image_path = sending_step.output_image_path
            else:
                super().set_input_as_output_from(sending_step)
    
    def infer_outputs_from_inputs(self, out_dir: str, der_type: str, suffix: str=None, ext: str=None, **extra_desc):
        """
        Infers output directory and prefix for TACs based on the input image path.

        Args:
            out_dir (str): Directory where the outputs will be saved.
            der_type (str): Type of derivatives.
            suffix (str, optional): Suffix for the output files. Defaults to None.
            ext (str, optional): Extension for the output files. Defaults to None.
            **extra_desc: Additional descriptive parameters.
        """
        sub_id, ses_id = parse_path_to_get_subject_and_session_id(self.input_image_path)
        outpath = gen_bids_like_dir_path(sub_id=sub_id, ses_id=ses_id, sup_dir=out_dir, modality='tacs')
        self.out_tacs_dir = outpath
        step_name_in_camel_case = snake_to_camel_case(self.name)
        self.out_tacs_prefix = f'sub-{sub_id}_ses-{ses_id}_desc-{step_name_in_camel_case}'
    
    @classmethod
    def default_write_tacs_from_segmentation_rois(cls, **overrides):
        """
        Provides a class method to create an instance with default parameters. All paths
        are set to empty strings, and `verbose=False`.

        Returns:
            TACsFromSegmentationStep: A new instance with default parameters.
        """

        defaults = dict(input_image_path='', segmentation_image_path='', segmentation_label_map_path='',
                        out_tacs_dir='', out_tacs_prefix='', verbose=False)
        override_dict = defaults | overrides

        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f'Invalid override: {err}. Using default instance instead.', stacklevel=2)
            return cls(**defaults)
    

class ResampleBloodTACStep(FunctionBasedStep):
    """
    A step in a processing pipeline for resampling blood Time Activity Curves (TACs) based on PET image timings.

    This class facilitates the resampling of blood TACs to match the scanner's time frames, providing properties and
    methods for handling the input and output paths, as well as the resampling thresholds. This class uses
    :func:`resample_blood_data_on_scanner_times<petpal.input_function.blood_input.resample_blood_data_on_scanner_times>`
    to perform the resampling.

    Attributes:
        raw_blood_tac_path (str): Path to the input raw blood TAC file.
        input_image_path (str): Path to the 4D-PET image file.
        resampled_tac_path (str): Path where the resampled TAC will be saved.
        lin_fit_thresh_in_mins (float): Threshold in minutes for linear fitting.

    """
    def __init__(self,
                 input_raw_blood_tac_path: str,
                 input_image_path: str,
                 out_tac_path: str,
                 lin_fit_thresh_in_mins=30.0,
                 rescale_constant:float = 37000.0):
        """
        Initializes a ResampleBloodTACStep with specified parameters.

        Args:
            input_raw_blood_tac_path (str): Path to the input raw blood TAC file.
            input_image_path (str): Path to the PET image file.
            out_tac_path (str): Path where the resampled TAC will be saved.
            lin_fit_thresh_in_mins (float): Threshold in minutes for linear fitting.
        """
        super().__init__(name='resample_PTAC_on_scanner',
                         function=blood_input.resample_blood_data_on_scanner_times,
                         blood_tac_path=input_raw_blood_tac_path,
                         reference_4dpet_img_path=input_image_path,
                         out_tac_path=out_tac_path,
                         lin_fit_thresh_in_mins=lin_fit_thresh_in_mins,
                         rescale_constant=rescale_constant)
        self._raw_blood_tac_path = input_raw_blood_tac_path
        self._input_image_path = input_image_path
        self._resampled_tac_path = out_tac_path
        self.lin_fit_thresh_in_mins = lin_fit_thresh_in_mins
        self.rescale_constant = rescale_constant
    
    def __repr__(self):
        """
        Provides an unambiguous string representation of the ResampleBloodTACStep instance.

        Returns:
            str: A string representation showing how the instance can be recreated.
        """
        cls_name = type(self).__name__
        info_str = [f'{cls_name}(']
        
        in_kwargs = ArgsDict(
                dict(input_raw_blood_tac_path=self.raw_blood_tac_path, input_image_path=self.input_image_path,
                     out_tac_path=self.resampled_tac_path, lin_fit_thresh_in_mins=self.lin_fit_thresh_in_mins))
        
        for arg_name, arg_val in in_kwargs.items():
            info_str.append(f'{arg_name}={repr(arg_val)},')
        info_str.append(')')
        
        return f'\n    '.join(info_str)
    
    @property
    def raw_blood_tac_path(self):
        """
        Gets the path to the input raw blood TAC file.

        Returns:
            str: The path to the input raw blood TAC file.
        """
        return self._raw_blood_tac_path
    
    @raw_blood_tac_path.setter
    def raw_blood_tac_path(self, raw_blood_tac_path):
        """
        Sets the path to the input raw blood TAC file and updates the function arguments.

        Args:
            raw_blood_tac_path (str): The new path to the raw blood TAC file.
        """
        self.kwargs['blood_tac_path'] = raw_blood_tac_path
        self._raw_blood_tac_path = raw_blood_tac_path
    
    @property
    def input_image_path(self):
        """
        Gets the path to the PET image file.

        Returns:
            str: The path to the PET image file.
        """
        return self._input_image_path
    
    @input_image_path.setter
    def input_image_path(self, input_image_path: str):
        """
        Sets the path to the PET image file and updates the function arguments.

        Args:
            input_image_path (str): The new path to the PET image file.
        """
        self.kwargs['reference_4dpet_img_path'] = input_image_path
        self._input_image_path = input_image_path
    
    @property
    def resampled_tac_path(self):
        """
        Gets the path where the resampled TAC will be saved.

        Returns:
            str: The path where the resampled TAC will be saved.
        """
        return self._resampled_tac_path
    
    @resampled_tac_path.setter
    def resampled_tac_path(self, resampled_tac_path):
        """
        Sets the path where the resampled TAC will be saved and updates the function arguments.

        Args:
            resampled_tac_path (str): The new path where the resampled TAC will be saved.
        """
        self.kwargs['out_tac_path'] = resampled_tac_path
        self._resampled_tac_path = resampled_tac_path
    
    def set_input_as_output_from(self, *sending_steps):
        """
        Sets the input image path based on the output from a specified sending step.

        Args:
            sending_step: The step from which to derive the input image path.
        """
        assert len(sending_steps) == 1, "ResampleBloodTACStep should only receive a single sending step."
        if isinstance(sending_steps[0], ImageToImageStep):
            self.input_image_path = sending_steps[0].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[0])
    
    def infer_outputs_from_inputs(self, out_dir: str, der_type, suffix='blood', ext='.tsv', **extra_desc):
        """
        Infers the output file path for resampled TAC based on the input raw blood TAC path.

        Args:
            out_dir (str): Directory where the outputs will be saved.
            der_type: Type of derivatives.
            suffix (str, optional): Suffix for the output files. Defaults to 'blood'.
            ext (str, optional): Extension for the output files. Defaults to '.tsv'.
            **extra_desc: Additional descriptive parameters.
        """
        sub_id, ses_id = parse_path_to_get_subject_and_session_id(self.raw_blood_tac_path)
        filepath = gen_bids_like_filepath(sub_id=sub_id, ses_id=ses_id, bids_dir=out_dir, modality='preproc',
                                          suffix=suffix, ext=ext, desc='OnScannerFrameTimes')
        self.resampled_tac_path = filepath
    
    @classmethod
    def default_resample_blood_tac_on_scanner_times(cls):
        """
        Provides a class method to create an instance with default parameters. All paths are set to empty-strings
        and the fitting threshold is set to 30 minutes.

        Returns:
            ResampleBloodTACStep: A new instance with default parameters.
        """
        return cls(input_raw_blood_tac_path='', input_image_path='', out_tac_path='', lin_fit_thresh_in_mins=30.0)


class ImageToImageStep(FunctionBasedStep):
    """
    A step in a processing pipeline for processing and transforming image files. This class handles input and output
    image paths, executes image transformation functions, and provides methods for setting inputs from other steps
    and inferring output paths.
    
    .. attention::
    
       The passed function must have the following arguments order:
       ``func(input_image, output_image, *args, **kwargs)`` where ``input_image`` and ``output_image``
       can be named something else. The first argument must be an input image path, and the second argument
       must be an output image path.

    Attributes:
        input_image_path (str): Path to the input image file.
        output_image_path (str): Path to the output image file.
        
    See Also:
        - :meth:`default_threshold_cropping`
        - :meth:`default_moco_frames_above_mean`
        - :meth:`default_register_pet_to_t1`

    """
    def __init__(self,
                 name: str,
                 function: Callable,
                 input_image_path: str,
                 output_image_path: str,
                 *args,
                 **kwargs) -> None:
        """
        Initializes an ImageToImageStep with specified parameters.

        Args:
            name (str): The name of the step.
            function (Callable): The function to execute.
            input_image_path (str): Path to the input image file.
            output_image_path (str): Path to the output image file.
            *args: Additional positional arguments for the transformation function.
            **kwargs: Additional keyword arguments for the transformation function.
        
        Notes:
            The passed function must have the following arguments order:
            ``func(input_image, output_image, *args, **kwargs)`` where ``input_image`` and ``output_image``
            can be named something else. The first argument must be an input image path, and the second argument
            must be an output image path.
        
        """
        super().__init__(name, function, *(input_image_path, output_image_path, *args), **kwargs)
        self.input_image_path = copy.copy(self.args[0])
        self.output_image_path = copy.copy(self.args[1])
        self.args = self.args[2:]
    
    def execute(self, copy_meta_file: bool = True) -> None:
        """
        Executes the function and optionally copies meta-data information using
        :func:`safe_copy_meta<petpal.utils.image_io.safe_copy_meta>`.

        Args:
            copy_meta_file (bool): Whether to copy meta information from input to output image. Defaults to True.
            
        Notes:
            Function must have the following arguments order: ``(input_image_path, output_image_path, *args, **kwargs)``
            where ``input_image`` and ``output_image`` are abritrary names.
            
        """
        print(f"(Info): Executing {self.name}")
        self.function(self.input_image_path, self.output_image_path, *self.args, **self.kwargs)
        if copy_meta_file:
            safe_copy_meta(input_image_path=self.input_image_path, out_image_path=self.output_image_path)
        print(f"(Info): Finished {self.name}")
    
    def __str__(self):
        """
        Provides a string representation of the ImageToImageStep instance.

        Returns:
            str: A string representation of the instance.
        """
        io_dict = ArgsDict({'input_image_path': self.input_image_path,
                            'output_image_path': self.output_image_path
                            })
        sup_str_list = super().__str__().split('\n')
        args_ind = sup_str_list.index("Arguments Passed:")
        sup_str_list.insert(args_ind, f"Input & Output Paths:\n{io_dict}")
        def_args_ind = sup_str_list.index("Default Arguments:")
        sup_str_list.pop(def_args_ind + 1)
        if not hasattr(self.function, '__wrapped__'):
            sup_str_list.pop(def_args_ind + 1)
        
        return "\n".join(sup_str_list)
    
    def set_input_as_output_from(self, *sending_steps) -> None:
        """
        Sets the input image path based on the output from a specified sending step.

        Args:
            sending_step (FunctionBasedStep): The step from which to derive the input image path.
        """
        assert len(sending_steps) == 1, "ImageToImageStep must have 1 sending step."
        if isinstance(sending_steps[0], ImageToImageStep):
            self.input_image_path = sending_steps[0].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[0])
    
    def can_potentially_run(self):
        """
        Checks if the step can potentially run based on input and output paths. Checks if all path related
        arguments are non-empty strings.

        Returns:
            bool: True if the step can potentially run, False otherwise.
        """
        input_img_non_empty_str = False if self.input_image_path == '' else True
        output_img_non_empty_str = False if self.output_image_path == '' else True
        return super().can_potentially_run() and input_img_non_empty_str and output_img_non_empty_str
    
    def infer_outputs_from_inputs(self,
                                  out_dir: str,
                                  der_type='preproc',
                                  suffix: str = 'pet',
                                  ext: str = '.nii.gz',
                                  **extra_desc):
        """
        Infers the output file path based on the input image path and other parameters.

        Args:
            out_dir (str): Directory where the outputs will be saved.
            der_type (str): Type of derivatives. Defaults to 'preproc'.
            suffix (str, optional): Suffix for the output files. Defaults to 'pet'.
            ext (str, optional): Extension for the output files. Defaults to '.nii.gz'.
            **extra_desc: Additional descriptive parameters.
        """
        sub_id, ses_id = parse_path_to_get_subject_and_session_id(self.input_image_path)
        step_name_in_camel_case = snake_to_camel_case(self.name)
        filepath = gen_bids_like_filepath(sub_id=sub_id, ses_id=ses_id, suffix=suffix, bids_dir=out_dir,
                                          modality=der_type, ext=ext, desc=step_name_in_camel_case, **extra_desc)
        self.output_image_path = filepath
    
    @classmethod
    def default_threshold_cropping(cls, name: str = 'thresh_crop',  **overrides):
        """
        Creates a default instance for threshold cropping using :class:`SimpleAutoImageCropper<petpal.preproc.image_operations_4d.SimpleAutoImageCropper>`.
        All paths are empty-strings.

        Args:
            name (str): Name of the step. Defaults to 'thresh_crop'.
            **overrides: Override default parameters.

        Returns:
            ImageToImageStep: A new instance for threshold cropping.
        """
        defaults = dict(name=name, function=SimpleAutoImageCropper, input_image_path='',
                        output_image_path='', )
        override_dict = defaults | overrides
        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            return cls(**defaults)

    @classmethod
    def default_windowed_moco(cls, name: str = 'windowed_moco', verbose=False, **overrides):
        """
        Creates a default instance for motion correction of frames using a windowed strategy.
        See :func:`windowed_motion_corr_to_target<petpal.preproc.motion_corr.windowed_motion_corr_to_target>`
        for more details. All paths are empty-strings.

        Args:
            name (str): Name of the step. Defaults to 'windowed_moco'.
            verbose:
            **overrides:

        Returns:
            ImageToImageStep: A new instance for windowed motion correction of frames.
        """
        defaults = dict(name=name, function=windowed_motion_corr_to_target,
                        input_image_path='', output_image_path='',
                        motion_target_option='weighted_series_sum', window_duration=60.0,
                        verbose=verbose)
        override_dict = defaults | overrides
        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            return cls(**defaults)

    @classmethod
    def default_register_pet_to_t1(cls, name:str = 'register_pet_to_t1', reference_image_path='', verbose=False, **overrides):
        """
        Creates a default instance for registering PET to T1 image using :func:`register_pet<petpal.preproc.register.register_pet>`.
        All paths are empty-strings.

        Args:
            name (str): Name of the step. Defaults to 'register_pet_to_t1'
            reference_image_path (str): Path to the reference image.
            verbose (bool): Whether to run in verbose mode.
            **overrides: Override default parameters.

        Returns:
            ImageToImageStep: A new instance for registering PET to T1 image.

        """
        defaults = dict(name=name, function=register_pet, input_image_path='', output_image_path='',
                        reference_image_path=reference_image_path, motion_target_option='weighted_series_sum',
                        verbose=verbose)
        override_dict = defaults | overrides
        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            return cls(**defaults)

    @classmethod
    def default_rescale_image(cls, name: str = 'rescale_image', **overrides):
        r"""Creates a default step-instance for rescaling an image using :class:`rescale_image<petpal.preproc.image_operations_4d.rescale_image>`.

        The defaults for this step will divide the input image by 37000.0 which is usually done to go
        from kBq/ml to nCi/ml.

        Notes:
             The function :class:`rescale_image<petpal.preproc.image_operations_4d.rescale_image>` is wrapped using
             :func:`ANTsImageToANTsImage<petpal.utils.decorators.ANTsImageToANTsImage>` since steps require input and output
             paths.

        Args:
            name (str): Name of the step. Defaults to 'rescale_image'.
            **overrides:

        Returns:
            ImageToImageStep: A new step instance for rescaling the input image.
        """
        defaults = dict(name=name, function=ANTsImageToANTsImage(rescale_image),
                        input_image_path='', output_image_path='',
                        rescale_constant=37000.0, op='/')
        override_dict = defaults | overrides

        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            return cls(**defaults)

    @classmethod
    def default_warp_pet_to_atlas(cls, name: str = 'warp_pet_to_atlas', anat_image_path: str = '', atlas_image_path: str = '', **overrides):
        """Creates a default step instance for warping a PET image to an atlas using :func:`~petpal.preproc.register.warp_pet_to_atlas`.

        Note:
            The defaults for this step do not include the *anat_image_path* or *atlas_image_path*, as these will depend
            on the pipeline construction. These paths will need to be set before a pipeline will be able to run.

        Args:
            name (str): Name of the step. Defaults to 'warp_pet_to_atlas'.
            anat_image_path (str): Path to anatomical image used to compute transform to atlas space. Defaults to an
                empty string ''.
            atlas_image_path (str): Path to atlas image to which pet will be warped
                (indirectly, via the anatomical image).
            **overrides: Override default parameters

        Returns:
            ImageToImageStep:
                A new step instance for warping the pet image to atlas space.
        """
        defaults = dict(name=name, function=ANTsImageToANTsImage(warp_pet_to_atlas),
                        input_image_path='', output_image_path='', anat_image_path=anat_image_path,
                        atlas_image_path=atlas_image_path, )
        override_dict = defaults | overrides

        try:
            return cls(**override_dict)
        except RuntimeError as err:
            warnings.warn(f"Invalid override: {err}. Using default instance instead.", stacklevel=2)
            return cls(**defaults)

class ImagePairToArrayStep(FunctionBasedStep):
    """
    A step in a processing pipeline for transforming two input image files into an output array.

    This class extends the :class:`FunctionBasedStep<petpal.pipelines.steps_base.FunctionBasedStep>` and is designed
    for tasks that require combining or processing information from two image inputs and outputing an array e.g.
    getting the mean activity from a 4D-PET image given a voxel mask image. It handles input image paths, output
    array paths, and provides methods for setting inputs from connected pipeline steps or for inferring output
    array paths.

    .. attention::

       The passed function must have the following arguments order:
       ``func(input_image, second_image, output_array, *args, **kwargs)`` where ``input_image`` and ``output_image``
       can be named something else. The first argument must be an input image path, the second argument must be the
       path to the second image, and the third argument must be an output array path.

    Attributes:
        input_image_path (str): Path to the first input image file.
        second_image_path (str): Path to the second input image file.
        output_array_path (str): Path to the output array file.

    """
    def __init__(self,
                 name: str,
                 function: Callable,
                 input_image_path: str,
                 second_image_path: str,
                 output_array_path: str,
                 *args,
                 **kwargs):
        """
        Initializes an ImagePairToArrayStep with specified parameters.

        Args:
            name (str): The name of the step.
            function (Callable): The function that uses the two input images and outputs an array.
            input_image_path (str): Path to the first input image file.
            second_image_path (str): Path to the second input image file.
            output_array_path (str): Path to the output array file.
            *args: Additional positional arguments for the processing function.
            **kwargs: Additional keyword arguments for the processing function.

        Notes:
            The passed function (``func``) must have the following arguments order:
            ``func(input_image, second_image, output_array, *args, **kwargs)`` where ``input_image``, ``second_image``,
            and ``output_array`` can be named something else. The first argument must be an input image path,
            the second argument must be the path to the second image, and the third argument must be an output
            array path.

        """
        super().__init__(name, function,
                         *(input_image_path, output_array_path,
                           second_image_path, *args),
                         **kwargs)
        self.input_image_path = copy.copy(self.args[0])
        self.second_image_path = copy.copy(self.args[1])
        self.output_array_path = copy.copy(self.args[2])
        self.args = copy.copy(self.args[3:])

    def execute(self):
        """
        Executes the function to process the two input images into an output array.

        The specified function will be called with the paths to the two input images, the output array,
        any additional arguments, and keyword arguments.

        """
        print(f"(Info): Executing {self.name}")
        self.function(self.input_image_path,
                      self.second_image_path,
                      self.output_array_path,
                      *self.args,
                      **self.kwargs)
        print(f"(Info): Finished {self.name}")

    def set_input_as_output_from(self, *sending_steps) -> None:
        """
        Sets the input image paths based on the output paths from other steps in the pipeline.
        The first sending step will set the input image path, and the second sending step will
        set the second image path.

        Args:
            sending_steps (tuple[FunctionBasedStep]): Two pipeline steps whose outputs will be used
                as the input image path and second image input path.

        Raises:
            AssertionError: If the number of provided sending steps is not exactly two.
        """
        assert len(sending_steps) == 2, "ImagePairToArrayStep must have 2 sending ImageToImageStep steps."
        if isinstance(sending_steps[0], ImageToImageStep):
            self.input_image_path = sending_steps[0].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[0])
        if isinstance(sending_steps[1], ImageToImageStep):
            self.second_image_path = sending_steps[1].output_image_path
        else:
            super().set_input_as_output_from(sending_steps[1])

    def infer_outputs_from_inputs(self,
                                  out_dir: str,
                                  der_type: str = 'tacs',
                                  suffix: str = 'tac',
                                  ext: str = '.tsv',
                                  **extra_desc):
        """
        Infers the output array path based on the inputs and specified parameters.

        This method generates a BIDS-like derivatives filepath for the output based on the subject and
        session IDs extracted from the input image path.

        Args:
            out_dir (str): Directory where the output array will be saved.
            der_type (str, optional): Type of derivative. Will set the sub-directory in `out_dir`. Defaults to 'tacs'.
            suffix (str, optional): Suffix for the output filename. Defaults to 'tac'.
            ext (str, optional): File extension for the output file. Defaults to '.tsv'.
            **extra_desc: Additional descriptive parameters for the output filename.
        """
        sub_id, ses_id = parse_path_to_get_subject_and_session_id(self.input_image_path)
        step_name_in_camel_case = snake_to_camel_case(self.name)
        filepath = gen_bids_like_filepath(sub_id=sub_id, ses_id=ses_id, suffix=suffix, bids_dir=out_dir,
                                          modality=der_type, ext=ext, desc=step_name_in_camel_case, **extra_desc)
        self.output_array_path = filepath


PreprocStepType = Union[TACsFromSegmentationStep, ResampleBloodTACStep, ImageToImageStep, ImagePairToArrayStep]


"""Tools for running statistics on PET imaging data."""
from collections.abc import Callable
import numpy as np
import ants

from ..meta.label_maps import LabelMapLoader
from .dimension import check_physical_space_for_ants_image_pair

def mean_value_in_region(input_img: ants.ANTsImage,
                         seg_img: ants.ANTsImage,
                         mappings: int | list[int]) -> float:
    """Calculate the mean value in a 3D PET image over a region based on one or more integer
    mappings corresponding to regions in a segmentation image.
    
    Args:
        input_img (ants.ANTsImage): 3D PET image over which to calculate the mean.
        seg_img (ants.ANTsImage): Segmentation image in same space as `input_img`.
        mappings (int | list[int]): One or more mappings to mask input_image over.

    Returns:
        region_mean (float): Mean PET value over voxels in the regions corresponding to
            `mappings`."""
    region_mask = ants.mask_image(input_img, seg_img, level=mappings)
    region_arr = region_mask.numpy().flatten()
    region_arr_nonzero = region_arr.nonzero()
    voxel_arr = region_arr[region_arr_nonzero]
    return voxel_arr.mean()


class RegionalStats:
    """Run statistics on each region in a parametric 3D PET kinetic model or other image.
    
    Options:
        * RegionalStats.mean: Mean value within each region
        * RegionalStats.std: Standard devation of values within each region.
        * RegionalStats.nvox: Number of voxels in each region.
        * RegionalStats.max: Maximum value in each region.
        * RegionalStats.min: Minimum value in each region.
        * RegionalStats.median: Median value in each region.
        * RegionalStats.get_stats(stats_func): Get a generic statistic run on each region. Runs
          function `stats_func` on each region, which must take a 1D array as the only argument.

    Example:

        .. code-block:: python

            import numpy as np
            from petpal.utils.stats import RegionalStats
            from petpal.utils.image_io import write_dict_to_json

            # Set up class
            input_image_path = 'sub-001_ses-01_space-mpr_desc-SUVR_pet.nii.gz'
            segmentation_image_path = 'sub-001_ses-01_seg.nii.gz'
            region_stats_obj = RegionalStats(input_image_path = input_image_path,
                                             segmentation_image_path = segmentation_image_path,
                                             label_map_option = 'freesurfer')
            
            # Preset statistic: get mean in each region
            region_means = region_stats_obj.mean
            write_dict_to_json(region_means,'sub-001_ses-01_RegionMeanSUVR.json')

            # Create function to get 95th percentile value for each region
            def calc_95th_percentile(arr: np.ndarray):
                return np.percentile(arr,95)
            region_95th = region_stats_obj.get_stats(calc_95th_percentile)
            write_dict_to_json(region_95th,'sub-001_ses-01_Region95thPercentileSUVR.json')

    :ivar pet_img: 3D PET image on which to get statistics for each region.
    :ivar seg_img: Segmentation image in same space as `pet_img` defining regions.
    :ivar label_map: Dictionary that assigns labels to regions in `seg_img`."""
    def __init__(self,
                 input_image_path: str,
                 segmentation_image_path: str,
                 label_map_option: str | dict):
        self.pet_img = ants.image_read(input_image_path)
        self.seg_img = ants.image_read(segmentation_image_path)
        assert check_physical_space_for_ants_image_pair(self.pet_img, self.seg_img), (
            "input image and anatomical image must occupy the same physical space")
        self.label_map = LabelMapLoader(label_map_option=label_map_option).label_map

    def get_voxels(self, label: str) -> np.ndarray:
        """Get the voxel array for the selected label.
        
        Args:
            label (str): Name of the region from which to extract voxels.
            
        Returns:
            voxel_arr (np.ndarray): Voxels in region as a flattened array."""
        mappings = self.label_map[label]
        region_mask = ants.mask_image(self.pet_img, self.seg_img, level=mappings)
        region_arr = region_mask.numpy().flatten()
        region_arr_nonzero = region_arr.nonzero()
        voxel_arr = region_arr[region_arr_nonzero]
        return voxel_arr

    def get_stats(self, stats_func: Callable, dtype: object=float) -> dict:
        """Get stats for all regions. Applies `stats_func` to the `voxel_arr` to return a
        value for each region. Set `dtype` depending on what type the output is (e.g. float or
        int).
        
        Args:
            stats_func (Callable): The function to run on each region's voxels. Must take a 1D
              array as the only required positional argument and return a value or np.ndarray.
            dtype (object): The dtype that `stats_func` returns, if not float. Typically float or
              int. Default float.
        
        Returns:
            region_stats (dict): The statistics for each region of interest.
        """
        region_stats = {}
        for label in self.label_map:
            voxel_arr = self.get_voxels(label=label)
            region_stat = stats_func(voxel_arr)
            if isinstance(region_stat,np.ndarray):
                region_stats[label] = region_stat.astype(dtype).tolist()
            else:
                region_stats[label] = dtype(region_stat)
        return region_stats

    @property
    def mean(self) -> dict:
        """Get mean value for each region."""
        return self.get_stats(stats_func=np.mean)

    @property
    def std(self) -> dict:
        """Get standard deviation of values for each region."""
        return self.get_stats(stats_func=np.std)

    @property
    def nvox(self) -> dict:
        """Get number of voxels in each region."""
        return self.get_stats(stats_func=len, dtype=int)

    @property
    def max(self) -> dict:
        """Get maximum value in each region."""
        return self.get_stats(stats_func=np.max)

    @property
    def min(self) -> dict:
        """Get minimum value in each region."""
        return self.get_stats(stats_func=np.min)

    @property
    def median(self) -> dict:
        """Get median value in each region."""
        return self.get_stats(stats_func=np.median)



"""Run regional statistics on PET Images"""
import argparse
from ..utils.stats import RegionalStats
from ..utils.image_io import write_dict_to_json

def main():
    """Run selected stats and write to file."""
    prog_desc = "Run statistics on a PET image for each region and write results to JSON file."
    epilog = "petpal-pet-stats --input-image-path suvr.nii.gz --segmentation-path " \
             "aparc+aseg.nii.gz --label-map freesurfer --statistic mean --out-path mean_suvr.json"
    parser = argparse.ArgumentParser(prog="STATS CLI", description=prog_desc, epilog=epilog)
    parser.add_argument("-i",
                        "--input-image-path",
                        required=True,
                        help="Path to 3D PET parametric image, such as SUVR or Vt.")
    parser.add_argument("-s",
                        "--segmentation-path",
                        required=True,
                        help="Path to 3D segmentation image, such as aparc+aseg.")
    parser.add_argument("-l",
                        "--label-map",
                        required=True,
                        help="Label map for ROIs. Presets include freesurfer, "
                             "freesurfer_merge_lr, perlcyno, perlcyno_merge_lr.")
    parser.add_argument("-f",
                        "--statistic",
                        required=True,
                        help="The statistic to calculate.",
                        choices=['mean','std','median','max','min','nvox'])
    parser.add_argument("-o",
                        "--out-path",
                        required=True,
                        help="Path to save output as JSON file.")
    args = parser.parse_args()

    stats_obj = RegionalStats(input_image_path=args.input_image_path,
                              segmentation_image_path=args.segmentation_path,
                              label_map_option=args.label_map)
    stats_result = getattr(stats_obj, args.statistic.lower())
    write_dict_to_json(meta_data_dict=stats_result, out_path=args.out_path)



"""
This module provides classes for time-activity curve (TAC) interpolation for Positron Emission Tomography (PET)
data. It enables the resampling of data evenly with respect to time, which is particularly useful when performing
convolutions with respect to time. We use :py:class:`scipy.interpolate.interp1d` for the interpolation.

The module comprises two classes: :class:`EvenlyInterpolate` and :class:`EvenlyInterpolateWithMax`.

The :class:`EvenlyInterpolate` class takes in TAC times and values and a specified delta time, :math:`\\Delta t`, to
resample data by interpolating TACs evenly with respect to time.

The :class:`EvenlyInterpolateWithMax` class extends the functionality of :class:`EvenlyInterpolate` by modifying the
calculation of delta time, :math:`\\Delta t`, to explicitly sample the maximum value of the TAC.

Example:
    In the proceeding example, we shall use the two classes to resample a fake TAC. For better visualization, we shall
    plot the resampled TACs shifted in the y-direction.
        
        .. plot::
            :include-source:
            :caption: Evenly interpolating TACs can sometimes miss the maximum value of the provided TAC given the
                time-step. Therefore, we can explicitly include the maximum value by modifying the time-step.
            
            import numpy as np
            from petpal.kinetic_modeling.tac_interpolation import EvenlyInterpolate, EvenlyInterpolateWithMax
            import matplotlib.pyplot as plt
            
            # define some dummy TAC times and values
            tac_times_in_minutes = np.array([0., 1., 2.5, 4.1, 7., 15.0])
            tac_values = np.array([0., 0.8, 2., 1.5, 0.6, 0.0])
        
            # instantiate EvenlyInterpolate object and resample TAC (and add shift for better visualization)
            even_interp = EvenlyInterpolate(tac_times=tac_times_in_minutes, tac_values=tac_values+0.25, delta_time=1.0)
            resampled_tac = even_interp.get_resampled_tac()
        
            # instantiate EvenlyInterpolateWithMax object and resample TAC (and add shift for better visualization)
            even_interp_max = EvenlyInterpolateWithMax(tac_times=tac_times_in_minutes, tac_values=tac_values+0.5, samples_before_max=3)
            resampled_tac_max = even_interp_max.get_resampled_tac()
            
            # plot the TAC and the resampled TACs
            fig, ax = plt.subplots(1,1, constrained_layout=True, figsize=(8,4))
            plt.plot(tac_times_in_minutes, tac_values, 'ko--', label='Raw TAC', zorder=2)
            plt.plot(*resampled_tac, 'ro-', label='Evenly Resampled TAC', zorder=1)
            plt.plot(*resampled_tac_max, 'bo-', label='Evenly Resampled TAC w/ Max', zorder=0)
            ax.text(s='resampled TACS are \\nshifted for visual clarity',
                    x=0.95, y=0.95, ha='right', va='top', transform=ax.transAxes, fontsize=16)
            plt.xlabel('Time (s)', fontsize=16)
            plt.ylabel('TAC Value', fontsize=16)
            plt.legend(bbox_to_anchor=(1.0, 0.5), loc='center left')
            plt.show()
        

Note:
    This module utilises :py:class:`scipy.interpolate.interp1d` for linear interpolation. Ensure Scipy is installed for
    this package to function.

TODO:
    * Test if the classes work correctly if the max value of the TAC is at the first or last time-point.

"""

import numpy as np
from scipy.interpolate import interp1d as sp_interpolation


class EvenlyInterpolate:
    """A class for basic evenly interpolating TACs with respect to time
    
    When performing convolutions with respect to time, care needs to be taken to account for the time-step between
    samples. One way to circumvent this problem is to resample data evenly with respect to the independent variable,
    or time.

    Uses :py:class:`scipy.interpolate.interp1d` to perform linear interpolation.

    Attributes:
        interp_func (scipy.interpolate.interp1d): Interpolation function given the provided TAC.
        resample_times (np.ndarray): Array containing evenly spaced TAC times.
        resample_vals (np.ndarray): Interpolated activities at the calculated resample times.
    """
    def __init__(self, tac_times: np.ndarray, tac_values: np.ndarray, delta_time: float) -> None:
        r"""Initializes an instance of the EvenlyInterpolate class for TAC interpolation.

        The constructor takes the Time-Activity Curve (TAC) times and values as inputs, along with a delta time
        value. It utilizes the SciPy function `scipy.interpolate.interp1d` to perform a linear interpolation of the
        provided TAC.

        After initializing, it generates an interpolation function based on the TAC times and values. It uses this function
        to calculate interpolated activities for a new array of evenly spaced times. These times start at the
        start time of the original TAC and end at the end time, with steps of delta time.


        Args:
            tac_times (np.ndarray): The time-points at which the original TAC activities were sampled.
                It should be a 1D numpy array with increasing float values.
            tac_values (np.ndarray): The corresponding activity values of the provided TAC sampled at tac_times.
                It should be a 1D numpy array of float values.
            delta_time (float): The time-step to use for the creation of evenly spaced resample times. It should be
                a positive float value.

        """
        self._tac_times = tac_times
        self._tac_values = tac_values
        self.interp_func = sp_interpolation(x=self._tac_times, y=self._tac_values, kind='linear', bounds_error=True)
        self.resample_times = np.arange(tac_times[0], tac_times[-1] + delta_time, delta_time)
        self.resample_vals = self.interp_func(self.resample_times)

    def get_resampled_tac(self) -> np.ndarray:
        """
        Returns the resampled times and values of the Time-Activity Curve (TAC).

        The function combines the resampled times and values into a single numpy array.

        Returns:
            (np.ndarray): An array containing two numpy arrays. The first array corresponds to the
            resampled times and the second array corresponds to the resampled activity values of the TAC.
            
        """
        return np.asarray([self.resample_times, self.resample_vals])
    
    
class EvenlyInterpolateWithMax(EvenlyInterpolate):
    r"""A class, extends :class:`EvenlyInterpolate`, and modifies the :math:`\Delta t` calculation such that the
    maximum value of the provided input TAC is explicitly sampled.
    
    Attributes:
        interp_func (scipy.interpolate.interp1d): Interpolation function given the provided TAC.
        resample_times (np.ndarray): Array containing evenly spaced TAC times.
        resample_vals (np.ndarray): Interpolated activities at the calculated resample times.
        dt (float): The :math:`\Delta t` for the resampled times such that the maximum value is explicitly sampled.
        
    See Also:
        :class:`EvenlyInterpolate`
        
    """
    def __init__(self, tac_times: np.ndarray, tac_values: np.ndarray, samples_before_max: float = 10.0):
        r"""Initializes an instance of the EvenlyInterpolateWithMax class for TAC interpolation.

        This subclass of the :class:EvenlyInterpolate class strives to ensure the maximum value of the Time Activity
        Curve (TAC) is included in the resampled TAC. It achieves this by determining a dynamic delta time value based
        on the input TAC and the user-specified number of samples before the TAC's max value.

        The constructor takes the TAC times and values as input, alongside a parameter that defines the number of
        samples to take before the max value of the TAC. It computes the delta time using a static method of the class,
        feeding that along with the original times and values into the parent class initializer to perform the actual
        interpolation.

        Args:
            tac_times (np.ndarray): The time-points at which the original TAC activities were sampled.
                It should be a 1D numpy array with increasing float values.
            tac_values (np.ndarray): The corresponding activity values of the provided TAC sampled at tac_times.
                It should be a 1D numpy array of float values.
            samples_before_max (float): The number of samples desired before the maximum TAC value is reached.
                It defaults to 10 if not specified and should be a positive float number.

        """
        self.dt = self.calculate_dt_for_even_spacing_with_max_sampled(tac_times, tac_values, samples_before_max)
        super().__init__(tac_times, tac_values, self.dt)
    
    @staticmethod
    def calculate_dt_for_even_spacing_with_max_sampled(tac_times: np.ndarray,
                                                       tac_values: np.ndarray,
                                                       samples_before_max: float) -> float:
        r"""Calculates the time-step or :math:`\Delta t` for evenly sampling the Time-Activity Curve (TAC)
        while preserving the maximum value of the TAC in the sample.

        This method determines the time-step so that when the TAC gets resampled,
        it is guaranteed that the maximum value of the TAC in the input is also sampled in the resample. The
        :math:`\Delta t` is calculated based on the equation:
        
        .. math::
        
            \Delta t = \frac{t_{\text{max}} - t_0}{N}
        
        where :math:`t_{\text{max}}` is the time at which the max value of the TAC is reached,
        :math:`t_0` is the start time, and :math:`N` is the number of samples desired before the max TAC value.

        Args:
            tac_times (np.ndarray): A 1D numpy array of float values, representing times at which TAC activities are
                sampled.
            tac_values (np.ndarray): A 1D numpy array of float values, indicating the activities of the given TAC at
                corresponding times.
            samples_before_max (float): The number of samples that you want before the maximum TAC value is reached.

        Returns:
            (float): The calculated time-step, :math:`\Delta t`, that ensures that the TAC is evenly sampled, and the
            maximum TAC value is included in the samples.
        """
        t_start = tac_times[0]
        t_for_max_val = tac_times[np.argmax(tac_values)]
        dt = (t_for_max_val - t_start) / samples_before_max
        return dt



"""
Module to run partial volume correction PET images using the Symmetric Geometric Transfer Matrix (sGTM) method. The
output behavior depends on the dimensionality of the input image. If the provided image is 3D, like a parametric SUV
image, the output will be one table (.tsv) with each unique ROI partial volume corrected per row. If the provided image
is 4D, the output will be a partial volume corrected TAC for each ROI.
"""
import os
import warnings
import numpy as np
from scipy.ndimage import gaussian_filter
import ants
import pandas as pd

from ..meta.label_maps import LabelMapLoader
from ..utils.dimension import check_physical_space_for_ants_image_pair
from ..utils.scan_timing import ScanTimingInfo
from ..utils.time_activity_curve import TimeActivityCurve
from ..preproc.segmentation_tools import unique_segmentation_labels

class Sgtm:
    """Handle sGTM partial volume correction on provided PET images.
    """
    def __init__(self,
                 input_image_path: str,
                 segmentation_image_path: str,
                 fwhm: float | tuple[float, float, float],
                 label_map_option: str | None = None,
                 zeroth_roi: bool = False):
        r"""Initialize running sGTM

        Args:
            input_image_path (str): Path to input parametric image on which sGTM will be run.
            segmentation_image_path (str): Path to segmentation image to which parametric image is
                aligned which is used to delineate regions for PVC.
            fwhm (float | tuple[float, float, float]): Full width at half maximum of the Gaussian 
                blurring kernel for each dimension in mm. If only one number is provided, it is used for all dimensions.
            label_map_option (Optional, str): Path to segmentation label map table,
                to be read by :func:`~.read_label_map_tsv`, for segmentation labeling ROIs in the outputs.
                Defaults to None.
            zeroth_roi (bool): If False, ignores the zeroth ``0`` label in calculations, often used to
                exclude background or non-ROI regions. Defaults to False.


        Example:

            .. code-block:: python

                import numpy as np
                from petpal.preproc.symmetric_geometric_transfer_matrix import Sgtm

                # Get 3D imaging and set FWHM parameter
                input_3d_image_path = "sub-001_ses-01_space-mpr_desc-SUV_pet.nii.gz"
                segmentation_image_path = "sub-001_ses-01_space-mpr_seg.nii.gz"
                fwhm = (5.,5.,5.)

                # initiate Sgtm class, run analysis, and save to an output TSV file.
                sgtm_analysis = Sgtm(input_image_path=input_3d_image_path,
                                     segmentation_image_path=segmentation_image_path,
                                     fwhm=fwhm,
                                     zeroth_roi = False)
                sgtm_analysis(output_path="sub-001_ses-01_pvc-sGTM_desc-SUV_pet.tsv")

                # Do the same with a time series 4D image. This results in a TAC for each region in
                # the segmentation file, that has been partial volume corrected with the sGTM
                # method.
                input_4d_image_path = "sub-001_ses-01_space-mpr_pet.nii.gz"
                sgtm_4d_analysis = Sgtm(input_image_path=input_4d_image_path,
                                        segmentation_image_path=segmentation_image_path,
                                        fwhm=fwhm,
                                        label_map_option=label_map_option,
                                        zeroth_roi = False)
                sgtm_4d_analysis(output_path="/path/to/output/directory/", out_tac_prefix='sub-001_ses-001_desc-sGTM')

        """
        self.input_image_path = input_image_path
        self.input_image = ants.image_read(input_image_path)
        self.segmentation_image = ants.image_read(segmentation_image_path)
        self.label_map_option = label_map_option
        self.fwhm = fwhm
        self.zeroth_roi = zeroth_roi
        self.sgtm_result = None

    def run(self):
        r"""Determine whether input image is 3D or 4D and run the correct sGTM method.

        If input image is 3D, implied usage is getting the average sGTM value for each region in
        the volume. If input image is 4D, implied usage is getting a time series average value for
        each frame in image within each region.
        """
        if self.input_image.dimension == 3:
            self.sgtm_result = self.run_sgtm_3d()

        elif self.input_image.dimension == 4:
            self.sgtm_result = self.run_sgtm_4d()

    def save(self, output_path: str, out_tac_prefix: str | None = None, one_tsv_per_region: bool = False):
        r"""Save sGTM results by writing the resulting array to one or more files.

        The behavior depends on the input image provided. If input image is 3D, saves the average sGTM value for each
        region in a TSV with one row per region. If input image is 4D, saves time series average values for each frame
        within each region. 4D operation saves a single file unless `one_tsv_per_region` is set to True.

        Args:
            output_path (str): Path to save sGTM results. For 3D images, this should typically be
                the full path to a .tsv file. For 4D images, this is the directory where the sGTM
                TACs will be saved.
            out_tac_prefix (Optional, str): Prefix of the TAC files. Typically, something like
                ``'sub-001_ses-001_desc-sGTM'``. Defaults to None.
            one_tsv_per_region (bool): If True, saves one tsv file for each unique region, as
                opposed to one file containing all TACs if False. Default False.
        """
        if self.input_image.dimension == 3:
            self.save_results_3d(sgtm_result=self.sgtm_result, out_tsv_path=output_path)
        elif self.input_image.dimension == 4:
            if one_tsv_per_region:
                self.save_results_4d_tacs(sgtm_result=self.sgtm_result,
                                          out_tac_dir=output_path,
                                          out_tac_prefix=out_tac_prefix)
            else:
                self.save_results_4d_multitacs(sgtm_result=self.sgtm_result,
                                               out_tac_dir=output_path,
                                               out_tac_prefix=out_tac_prefix)

    def __call__(self, output_path: str, out_tac_prefix: str | None = None):
        r"""Run sGTM and save results.

        Applies :meth:`run_sgtm` for 3D images and :meth:`run_sgtm_4d`
        for 4D images.

        Args:
            output_path (str): Path to save sGTM results. For 3D images, this is a .tsv file. For
                4D images, this is a directory.
        """
        self.run()
        self.save(output_path=output_path, out_tac_prefix=out_tac_prefix)

    @property
    def sigma(self) -> list[float]:
        r"""Blurring kernal sigma for sGTM based on the input FWHM.

        Returns:
            sigma (list[float]): List of sigma blurring radii for Gaussian kernel. Each sigma value
                corresponds to an axis: x, y, and z. Values are determined based on the FWHM input
                to the object and the voxel dimension in the input image.
        """
        resolution = self.segmentation_image.spacing
        if isinstance(self.fwhm, (float, int)):
            sigma = [(self.fwhm / 2.355) / res for res in resolution]
        else:
            sigma = [(fwhm_i / 2.355) / res_i for fwhm_i, res_i in zip(self.fwhm, resolution)]
        return sigma

    @property
    def unique_labels(self) -> tuple[np.ndarray, list[str]]:
        r"""Get unique ROI indices and corresponding labels. If a segmentation label map was provided at instantiation,
        the table will be used to determine the unique ROI indices and labels. If a segmentation label map was not
        provided at instantiation, the segmentation image will be used to determine the unique ROI indices, and labels
        will be inferred in numerical order of indices.

        Returns:
            unique_segmentation_labels (tuple[np.ndarray, list[str]]): Tuple containing the unique ROI indices (mapping)
                and inferred segmentation labels.
        """
        if self.label_map_option is None:
            region_index_map = unique_segmentation_labels(segmentation_img=self.segmentation_image,
                                                          zeroth_roi=self.zeroth_roi)
            region_short_names = [f'UNK{i:05d}' for i in region_index_map]
        else:
            warnings.warn(f"Using label map for sGTM with option: {self.label_map_option}. Users "
                          "should be aware that any label map used with sGTM should contain a "
                          "complete list of regions for the brain. Review label map and "
                          "segmentation to ensure this criteria is met, or use sGTM without "
                          "label map for automated complete region mapping.")
            seg_label_map = LabelMapLoader(label_map_option=self.label_map_option).label_map
            unique_mappings = unique_segmentation_labels(segmentation_img=self.segmentation_image,
                                                       zeroth_roi=self.zeroth_roi)
            region_index_map = []
            region_short_names = []
            label_map_labels = list(seg_label_map.keys())
            label_map_mappings = list(seg_label_map.values())
            for mapping in unique_mappings:
                if mapping in label_map_mappings:
                    id_mapping_index = label_map_mappings.index(mapping)
                    region_index_map.append(label_map_mappings[id_mapping_index])
                    region_short_names.append(label_map_labels[id_mapping_index])
                else:
                    region_index_map.append(mapping)
                    region_short_names.append(f'UNK{mapping:05d}')
        return (region_index_map, region_short_names)


    @staticmethod
    def get_omega_matrix(voxel_by_roi_matrix: np.ndarray) -> np.ndarray:
        r"""Get the Omega matrix for sGTM. See :meth:`run_sgtm` for details.

        Args:
            voxel_by_roi_matrix (np.ndarray): The ``V`` matrix described in :meth:`run_sgtm`
                obtained by applying a Gaussian filter to each ROI.

        Returns:
            omega (np.ndarray): ``\Omega`` matrix as described in :meth:`run_sgtm`.
        """
        omega = voxel_by_roi_matrix.T @ voxel_by_roi_matrix
        return omega


    @staticmethod
    def solve_sgtm(omega: np.ndarray,
                   voxel_by_roi_matrix: np.ndarray,
                   input_numpy: np.ndarray) -> tuple:
        r"""Set up and solve linear equation for sGTM.

        Args:
            omega (np.ndarray): The Omega matrix for sGTM. See :meth:`run_sgtm` for details.
            voxel_by_roi_matrix (np.ndarray): The ``V`` matrix for sGTM. See :meth:`run_sgtm`
                for more details.
            input_numpy (np.ndarray): The input 3D PET image converted to numpy array.
        """
        t_vector = voxel_by_roi_matrix.T @ input_numpy.ravel()
        t_corrected = np.linalg.solve(omega, t_vector)
        condition_number = np.linalg.cond(omega)

        return t_corrected, condition_number


    @staticmethod
    def get_voxel_by_roi_matrix(unique_labels: np.ndarray,
                                segmentation_arr: np.ndarray,
                                sigma: list[float]) -> np.ndarray:
        r"""Get the ``V`` matrix for sGTM by blurring each ROI and converting into vectors.
        See :meth:`run_sgtm` for more details.

        Args:
            unique_labels (np.ndarray): Array containing unique values in the discrete segmentation
                image.
            segmentation_arr (np.ndarray): Array containing discrete segmentation image converted
                to a numpy array.
            sigma (list[float]): List of sigma blurring radii on x, y, z axes respectively.

        Returns:
            voxel_by_roi_matrix (np.ndarray): The blurred ROI matrix for sGTM.
        """
        voxel_by_roi_matrix = np.zeros((segmentation_arr.size, len(unique_labels)))

        for i, label in enumerate(unique_labels):
            masked_roi = (segmentation_arr == label).astype('float32')
            blurred_roi = gaussian_filter(masked_roi, sigma=sigma)
            voxel_by_roi_matrix[:, i] = blurred_roi.ravel()

        return voxel_by_roi_matrix.astype(np.float32)


    def run_sgtm_3d(self) -> tuple[np.ndarray, np.ndarray, float]:
        r"""Apply Symmetric Geometric Transfer Matrix (SGTM) method for Partial Volume Correction
        (PVC) to PET images based on ROI labels.

        This method involves using a matrix-based approach to adjust the PET signal intensities for
        the effects of partial volume averaging.

        Returns:
            Tuple[np.ndarray, np.ndarray, float]:
                - np.ndarray: Array of unique ROI labels.
                - np.ndarray: Corrected PET values after applying PVC.
                - float: Condition number of the omega matrix, indicating the numerical stability
                    of the inversion.

        Raises:
            AssertionError: If `self.input_image` and `self.segmentation_image` do not have the
                same dimensions.

        Notes:
            The SGTM method uses the matrix :math:`\Omega` (omega), defined as:

            .. math::
            
                \Omega = V^T V

            where :math:`V` is the matrix obtained by applying Gaussian filtering to each ROI,
            converting each ROI into a vector. The element :math:`\Omega_{ij}` of the matrix
            :math:`\Omega` is the dot product of vectors corresponding to the i-th and j-th ROIs,
            representing the spatial overlap between these ROIs after blurring.

            The vector :math:`t` is calculated as:

            .. math::
            
                t = V^T p

            where :math:`p` is the vectorized PET image. The corrected values,
            :math:`t_{corrected}`, are then obtained by solving the linear system:

            .. math::
            
                \Omega t_{corrected} = t

            This provides the estimated activity concentrations corrected for partial volume
            effects in each ROI.
        """
        if self.input_image.shape != self.segmentation_image.shape:
            raise AssertionError("PET and ROI images must be the same dimensions")
        input_numpy = self.input_image.numpy()
        segmentation_arr = self.segmentation_image.numpy()

        unique_labels = self.unique_labels[0]

        voxel_by_roi_matrix = Sgtm.get_voxel_by_roi_matrix(unique_labels=unique_labels,
                                                           segmentation_arr=segmentation_arr,
                                                           sigma=self.sigma)
        omega = Sgtm.get_omega_matrix(voxel_by_roi_matrix=voxel_by_roi_matrix)
        t_corrected, condition_number = Sgtm.solve_sgtm(omega=omega,
                                                        voxel_by_roi_matrix=voxel_by_roi_matrix,
                                                        input_numpy=input_numpy)

        return unique_labels, t_corrected, condition_number

    def run_sgtm_4d(self) -> np.ndarray:
        r"""Calculated partial volume corrected TACs on a 4D image by running sGTM on each frame in
        the 4D image.
        
        This results in a time series of average activity for each region specified in the
        segmentation image. This can then be used for kinetic modeling.

        Returns:
            frame_results (np.ndarray): Average activity in each region calculated with sGTM
                for each frame.
        """
        if not check_physical_space_for_ants_image_pair(self.input_image,
                                                        self.segmentation_image):
            raise AssertionError("PET and ROI images must be the same dimensions")
        pet_frame_list = self.input_image.ndimage_to_list()
        segmentation_arr = self.segmentation_image.numpy()

        unique_labels = self.unique_labels[0]

        voxel_by_roi_matrix = Sgtm.get_voxel_by_roi_matrix(unique_labels=unique_labels,
                                                           segmentation_arr=segmentation_arr,
                                                           sigma=self.sigma)
        omega = Sgtm.get_omega_matrix(voxel_by_roi_matrix=voxel_by_roi_matrix)

        frame_results = []
        for frame in pet_frame_list:
            input_numpy = frame.numpy()
            t_corrected, _cond_num = Sgtm.solve_sgtm(omega=omega,
                                                     voxel_by_roi_matrix=voxel_by_roi_matrix,
                                                     input_numpy=input_numpy)
            frame_results += [t_corrected]

        return np.asarray(frame_results)

    def save_results_3d(self, sgtm_result: tuple, out_tsv_path: str):
        r"""Saves the result of an sGTM calculation.

        Result is saved as one value for each of the unique regions found in the segmentation
        image.

        Args:
            sgtm_result (tuple): Output of :meth:`run_sgtm_3d`
            out_tsv_path (str): File path to which results are saved.
        """
        sgtm_result_to_write = pd.DataFrame(columns=['Region','Mean'])
        sgtm_result_to_write['Region'] = self.unique_labels[1]
        sgtm_result_to_write['Mean'] = sgtm_result[1]
        sgtm_result_to_write.to_csv(out_tsv_path,sep='\t',index=False)

    def save_results_4d_tacs(self,
                             sgtm_result: np.ndarray,
                             out_tac_dir: str,
                             out_tac_prefix: str,):
        r"""Saves the result of an sGTM calculation on a 4D PET series.

        Result is saved as a TAC for each of the unique regions found in the segmentation image.

        Args:
            sgtm_result (np.ndarray): Array of results from :meth:`run_sgtm_4d`
            out_tac_dir (str): Path to folder where regional TACs will be saved.
            out_tac_prefix (str): Prefix of the TAC files.
        """
        os.makedirs(out_tac_dir, exist_ok=True)
        input_image_path = self.input_image_path
        tac_times = ScanTimingInfo.from_nifti(image_path=input_image_path).center_in_mins

        tac_array = np.asarray(sgtm_result).T

        for i, (_label, name) in enumerate(zip(*self.unique_labels)):
            pvc_tac = TimeActivityCurve(times=tac_times,
                                        activity=tac_array[i,:])
            out_tac_path = os.path.join(f'{out_tac_dir}', f'{out_tac_prefix}_seg-{name}_tac.tsv')
            pvc_tac.to_tsv(filename=out_tac_path)

    def save_results_4d_multitacs(self,
                                  sgtm_result: np.ndarray,
                                  out_tac_dir: str,
                                  out_tac_prefix: str):
        """Like :meth:`save_results_4d_tacs`, but saves all TACs to a single file.

        Args:
            sgtm_result (np.ndarray): Array of results from :meth:`run_sgtm_4d`
            out_tac_dir (str): Path to folder where regional TACs will be saved.
            out_tac_prefix (str): Prefix of the TAC files.
        """
        os.makedirs(out_tac_dir, exist_ok=True)
        input_image_path = self.input_image_path
        scan_timing = ScanTimingInfo.from_nifti(image_path=input_image_path)
        tac_time_starts = scan_timing.start_in_mins
        tac_time_ends = scan_timing.end_in_mins

        tac_array = np.asarray(sgtm_result).T
        tacs_data_columns = ['frame_start(min)','frame_end(min)']+self.unique_labels[1]
        tacs_data = pd.DataFrame(columns=tacs_data_columns)

        tacs_data['frame_start(min)'] = tac_time_starts
        tacs_data['frame_end(min)'] = tac_time_ends
        for i, (_label, name) in enumerate(zip(*self.unique_labels)):
            tacs_data[name] = tac_array[i,:]
            tacs_data[f'{name}_unc'] = np.full(tac_array.shape[1],np.nan)
        tacs_data.to_csv(f'{out_tac_dir}/{out_tac_prefix}_multitacs.tsv', sep='\t', index=False)
